# -*- coding: utf-8 -*-
#from sampy import *
#from SocketServer import ThreadingMixIn
import sampy
from gavi import logging as logging_
import threading

logger = logging_.getLogger("gavi.samp")


class Samp(object):
	def __init__(self, daemon=True, name=None):
		self.client = sampy.SAMPIntegratedClient(metadata = {"samp.name":"Gavi client" if name is None else name,
										"samp.description.text": "Gavi client" if name is None else name,
										"gavi.samp.version":"0.01"}, callable=True)


		# sampy doesn't make this thread Daeamon, so the python process never stops on the cmd line
		# this fixes that
		def _myrun_client():
			if self.client.client._callable:
				self.client.client._thread = threading.Thread(target = self.client.client._serve_forever)
				self.client.client._thread.setDaemon(True)
				self.client.client._thread.start()
		if daemon:
			self.client.client._run_client = _myrun_client
		connected = False
		try:
			self.client.connect()
			connected = True
		except sampy.SAMPHubError as e:
			print(("error connecting to hub", e))
		
		if connected:
			#self.client.client._thread.setDaemon(False)
			logger.info("connected to SAMP hub")
			logger.info("binding events")
			self.client.bindReceiveCall			("table.load.votable", self._onTableLoadVotable)
			self.client.bindReceiveNotification	("table.load.votable", self._onTableLoadVotable)
			#self.client.bindReceiveNotification	("table.highlight.row", self._onSampNotification)
			#self.client.bindReceiveMessage("table.load.votable", self._onSampCall)
			#self.client.bindReceiveResponse("table.load.votable", self._onSampCall)
			
			#self.client.bindReceiveCall("samp.*", self._onSampCall)
			#self.client.bindReceiveNotification("samp.*", self._onSampNotification)
			
			#self.client.bindReceiveCall("table.*", self._onSampCall)
			#self.client.bindReceiveNotification("table.*", self._onSampNotification)
			#self.client.bindReceiveMessage("table.*", self._onSampCall)
			#self.client.bindReceiveResponse("table.*", self._onSampCall)
			
			#self.client.bindReceiveMessage("table.votable.*", self._onSampCall)
			#self.client.bindReceiveResponse("table.votable.*", self._onSampCall)
			
	#def connect(self):
	#	self.client.connect()
		self.tableLoadCallbacks = []
	
	def _onTableLoadVotable(self, private_key, sender_id, msg_id, mtype, params, extra):
		print(("Msg:", repr(private_key), repr(sender_id), repr(msg_id), repr(mtype), repr(params), repr(extra)))
		try:
			url = params["url"]
			table_id = params["table-id"]
			name = params["name"]
			for callback in self.tableLoadCallbacks:
				callback(url, table_id, name)
		except:
			logger.exception("event handler failed")
		
		if msg_id != None: # if SAMP call, send a reply
			self.client.ereply(msg_id, sampy.SAMP_STATUS_OK, result = {"txt": "loaded"})
		
	def _onSampNotification(self, private_key, sender_id, mtype, params, extra):
		print(("Notification:", repr(private_key), repr(sender_id), repr(mtype), repr(params), repr(extra)))
		
	def _onSampCall(self, private_key, sender_id, msg_id, mtype, params, extra):
		print("----")
		try:
			print(("Call:", repr(private_key), repr(sender_id), repr(msg_id), repr(mtype), repr(params), repr(extra)))
			self.client.ereply(msg_id, sampy.SAMP_STATUS_OK, result = {"txt": "printed"})
		except:
			print("errrrrrrororrrr hans!")




from functools import reduce
import operator
import os
import sys
import logging

import numpy as np
import pyarrow as pa

import dask.base
from vaex.expression import Expression
from vaex.utils import _normalize_selection_name
from .expression import _unary_ops, _binary_ops, reversable


from .stat import _Statistic
from vaex import encoding
from .datatype import DataType
from .docstrings import docsubst
import vaex.utils


list_ = list # dangerous, we override list in this module (the aggregator)

logger = logging.getLogger("vaex.agg")

if vaex.utils.has_c_extension:
    import vaex.superagg

_min = min  # we're gonna overwrite builtin min
aggregates = {}


def register(f, name=None):
    name = name or f.__name__
    aggregates[name] = f
    return f


@encoding.register('aggregation')
class aggregation_encoding:
    @staticmethod
    def encode(encoding, agg):
        return agg.encode(encoding)

    @staticmethod
    def decode(encoding, agg_spec):
        agg_spec = agg_spec.copy()
        type = agg_spec.pop('aggregation')
        f = aggregates[type]
        args = []
        if type == '_sum_moment':
            if 'parameters' in agg_spec:  # renameing between spec and implementation
                agg_spec['moment'] = agg_spec.pop('parameters')[0]
        if 'expressions' in agg_spec:
            args = agg_spec.pop('expressions')
        if type == "list":
            if "parameters" in agg_spec:  # renameing between spec and implementation
                agg_spec["dropnan"], agg_spec["dropmissing"] = agg_spec.pop("parameters")
        return f(*args, **agg_spec)


class AggregatorDescriptor:
    def __repr__(self):
        args = [*self.expressions]
        return 'vaex.agg.{}({!r})'.format(self.short_name, ", ".join(map(str, args)))

    def pretty_name(self, id, df):
        if id is None:
            id = "_".join(map(lambda k: df[k]._label, self.expressions))
        return '{0}_{1}'.format(id, self.short_name)

    def finish(self, value):
        return value


class AggregatorExpressionUnary(AggregatorDescriptor):
    def __init__(self, name, op, code, agg):
        self.agg = agg
        self.name = name
        self.op = op
        self.code = code
        self.expressions = self.agg.expressions
        self.selection = self.agg.selection

    def __repr__(self):
        return f'{self.code}{self.agg!r}'

    @property
    def edges(self):
        return self.agg.edges

    @edges.setter
    def edges(self, value):
        self.agg.edges = value

    def add_tasks(self, df, binners, progress):
        tasks, result = self.agg.add_tasks(df, binners, progress)
        @vaex.delayed
        def finish(value):
            return self.finish(value)
        return tasks, finish(result)

    def finish(self, value):
        return self.op(value)


class AggregatorExpressionBinary(AggregatorDescriptor):
    def __init__(self, name, op, code, agg1, agg2, reverse=False):
        self.agg1 = agg1
        self.agg2 = agg2
        self.reverse = reverse
        self.name = name
        self.op = op
        self.code = code
        self.expressions = self.agg1.expressions + self.agg2.expressions
        self.selection = self.agg1.selection
        self.short_name = f'{self.code}{self.agg2.short_name}'
        if self.agg1.selection != self.agg2.selection:
            raise ValueError(f'Selections of aggregator for binary op {self.op} should be the same')

    def __repr__(self):
        if self.reverse:
            return f'({self.agg2!r} {self.code} {self.agg1!r})'
        else:
            return f'({self.agg1!r} {self.code} {self.agg2!r})'

    @property
    def edges(self):
        assert self.agg1.edges == self.agg2.edges
        return self.agg1.edges

    @edges.setter
    def edges(self, value):
        self.agg1.edges = value
        self.agg2.edges = value

    def add_tasks(self, df, binners, progress):
        progressbar = vaex.utils.progressbars(progress, title=repr(self))
        tasks1, result1 = self.agg1.add_tasks(df, binners, progress=progressbar)
        tasks2, result2 = self.agg2.add_tasks(df, binners, progress=progressbar)
        @vaex.delayed
        def finish(value1, value2):
            return self.finish(value1, value2)
        return tasks1 + tasks2, finish(result1, result2)

    def finish(self, value1, value2):
        if self.reverse:
            return self.op(value2, value1)
        return self.op(value1, value2)


class AggregatorExpressionBinaryScalar(AggregatorDescriptor):
    def __init__(self, name, op, code, agg, scalar, reverse=False):
        self.agg = agg
        self.scalar = scalar
        self.name = name
        self.code = code
        self.op = op
        self.reverse = reverse
        self.expressions = self.agg.expressions
        self.selection = self.agg.selection

    def __repr__(self):
        if self.reverse:
            return f'({self.scalar!r} {self.code} {self.agg!r})'
        else:
            return f'({self.agg!r} {self.code} {self.scalar!r})'

    @property
    def edges(self):
        return self.agg.edges

    @edges.setter
    def edges(self, value):
        self.agg.edges = value

    def add_tasks(self, df, binners, progress):
        progressbar = vaex.utils.progressbars(progress, title=repr(self))
        tasks, result = self.agg.add_tasks(df, binners, progress=progressbar)
        @vaex.delayed
        def finish(value):
            return self.finish(value)
        return tasks, finish(result)

    def finish(self, value):
        if self.reverse:
            return self.op(self.scalar, value)
        return self.op(value, self.scalar)


for op in _binary_ops:
    def wrap(op=op):
        def f(a, b):
            if isinstance(a, AggregatorDescriptor):
                if isinstance(b, AggregatorDescriptor):
                    return AggregatorExpressionBinary(op['name'], op['op'], op['code'], a, b)
                else:
                    return AggregatorExpressionBinaryScalar(op['name'], op['op'], op['code'], a, b)
            else:
                raise RuntimeError('Cannot happen')
        setattr(AggregatorDescriptor, '__%s__' % op['name'], f)
        if op['name'] in reversable:
            def f(a, b):
                if isinstance(a, AggregatorDescriptor):
                    if isinstance(b, AggregatorDescriptor):
                        raise RuntimeError('Cannot happen')
                    else:
                        return AggregatorExpressionBinaryScalar(op['name'], op['op'], op['code'], a, b, reverse=True)
                else:
                    raise RuntimeError('Cannot happen')
            setattr(AggregatorDescriptor, '__r%s__' % op['name'], f)
    wrap(op)
for op in _unary_ops:
    def wrap(op=op):
        def f(a):
            return AggregatorExpressionUnary(op['name'], op['op'], op['code'], a)
        setattr(AggregatorDescriptor, '__%s__' % op['name'], f)
    wrap(op)


class AggregatorDescriptorBasic(AggregatorDescriptor):
    def __init__(self, name, expressions, short_name, multi_args=False, agg_args=[], selection=None, edges=False):
        self.name = name
        self.short_name = short_name
        self.agg_args = agg_args
        self.edges = edges
        self.selection = _normalize_selection_name(selection)
        assert isinstance(expressions, (list_, tuple))
        for e in expressions:
            assert not isinstance(e, (list_, tuple))
        self.expressions = [str(k) for k in expressions]
        if len(self.expressions) == 1 and self.expressions[0] == '*':
            self.expressions = []

    def __repr__(self):
        args = [*self.expressions, *self.agg_args]
        return 'vaex.agg.{}({!r})'.format(self.short_name, ", ".join(map(str, args)))

    def encode(self, encoding):
        spec = {'aggregation': self.short_name}
        if len(self.expressions) == 0:
            pass
        else:
            spec['expressions'] = [str(k) for k in self.expressions]
        if self.selection is not None:
            spec['selection'] = str(self.selection) if isinstance(self.selection, Expression) else self.selection
        if self.edges:
            spec['edges'] = True
        if self.agg_args and self.short_name not in ['first', 'last']:
            spec['parameters'] = self.agg_args
        return spec

    def _prepare_types(self, df):
        if len(self.expressions) == 0 and self.short_name == "count":
            self.dtype_in = DataType(np.dtype('int64'))
            self.dtype_out = DataType(np.dtype('int64'))
        else:
            self.dtypes_in = [df[str(e)].data_type().index_type for e in self.expressions]
            self.dtype_in = self.dtypes_in[0]
            self.dtype_out = self.dtype_in
            if self.short_name == "count":
                self.dtype_out = DataType(np.dtype('int64'))
            if self.short_name in ['sum', 'summoment']:
                self.dtype_out = self.dtype_in.upcast()

    def add_tasks(self, df, binners, progress):
        progressbar = vaex.utils.progressbars(progress)
        self._prepare_types(df)
        task = vaex.tasks.TaskAggregation(df, binners, self)
        task = df.executor.schedule(task)
        progressbar.add_task(task, repr(self))
        @vaex.delayed
        def finish(value):
            return self.finish(value)
        return [task], finish(task)

    def _create_operation(self, grid, nthreads):
        if self.name in ["AggFirst", "AggList"]:
            if len(self.dtypes_in) == 1:
                # rows use int64
                agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtypes_in[0], vaex.dtype(np.dtype('int64')))
            else:
                agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtypes_in[0], self.dtypes_in[1])
        else:
            agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtype_in)
        if self.dtype_out.is_primitive or self.dtype_out.is_temporal:
            bytes_per_cell = self.dtype_out.numpy.itemsize
        else:
            bytes_per_cell = self.dtype_out.value_type.numpy.itemsize
        cells = reduce(operator.mul, [len(binner) for binner in grid.binners], 1)
        grids = nthreads
        ncells = len(grid)
        # the more grid cells we have, the less more work have to do to
        # merge/reduce the grids
        if ncells >= 1e4:
            grids = _min(32, nthreads)
        if ncells >= 1e5:
            grids = _min(16, nthreads)
        if ncells >= 1e6:
            grids = _min(8, nthreads)
        if grids < 1:
            grids = 1
        if logger.isEnabledFor(logging.INFO):
            logger.info("Using %r grids for %r thread for aggerator %r (total grid cells %s)", grids, nthreads, self, f"{ncells:,}")
        if self.short_name in ["list"]:
            # cannot predict memory usage
            predicted_memory_usage = None
            grids = 1
        else:
            predicted_memory_usage = bytes_per_cell * cells * grids
            vaex.memory.local.agg.pre_alloc(predicted_memory_usage, f"aggregator data for {agg_op_type}")
        agg_op = agg_op_type(grid, grids, nthreads, *self.agg_args)
        used_memory = sys.getsizeof(agg_op)

        if predicted_memory_usage is not None:
            if predicted_memory_usage != used_memory:
                raise RuntimeError(f'Wrong prediction for {agg_op_type}, expected to take {predicted_memory_usage} bytes but actually used {used_memory}')
        else:
            vaex.memory.local.agg.pre_alloc(used_memory, f"aggregator data for {agg_op_type}")
        return agg_op

    def get_result(self, agg_operation):
        grid = agg_operation.get_result()
        if not self.edges:
            def binner2slice(binner):
                if 'BinnerScalar_' in str(binner):
                    return slice(2, -1)
                elif 'BinnerOrdinal_' in str(binner):
                    return slice(0, -2)
                else:
                    raise TypeError(f'Binner not supported with edges=False {binner}')
            slices = [binner2slice(binner) for binner in agg_operation.grid.binners]
            grid = grid[tuple(slices)]
        return grid


class AggregatorDescriptorNUnique(AggregatorDescriptorBasic):
    def __init__(self, name, expression, short_name, dropmissing, dropnan, selection=None, edges=False):
        super(AggregatorDescriptorNUnique, self).__init__(name, expression, short_name, selection=selection, edges=edges)
        self.dropmissing = dropmissing
        self.dropnan = dropnan

    def encode(self, encoding):
        spec = super().encode(encoding)
        if self.dropmissing:
            spec['dropmissing'] = self.dropmissing
        if self.dropnan:
            spec['dropnan'] = self.dropnan
        return spec

    def _prepare_types(self, df):
        super()._prepare_types(df)
        self.dtype_out = DataType(np.dtype('int64'))

    def _create_operation(self, grid, nthreads):
        grids = 1  # this is using a shared hashmap, which is thread safe
        agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtype_in)
        cells = reduce(operator.mul, [len(binner) for binner in grid.binners], 1)
        grid0 = vaex.superagg.Grid([])
        agg_op_test = agg_op_type(grid0, grids, nthreads, self.dropmissing, self.dropnan)
        predicted_memory_usage = sys.getsizeof(agg_op_test) * cells
        vaex.memory.local.agg.pre_alloc(predicted_memory_usage, f"aggregator data for {agg_op_type}")
        agg_op = agg_op_type(grid, grids, nthreads, self.dropmissing, self.dropnan)
        used_memory = sys.getsizeof(agg_op)
        if predicted_memory_usage != used_memory:
            raise RuntimeError(f'Wrong prediction for {agg_op_type}, expected to take {predicted_memory_usage} bytes but actually used {used_memory}')

        return agg_op


class AggregatorDescriptorMulti(AggregatorDescriptor):
    """Uses multiple operations/aggregation to calculate the final aggretation"""
    def __init__(self, name, expressions, short_name, selection=None, edges=False):
        self.name = name
        self.short_name = short_name
        self.expressions = expressions
        self.selection = selection
        self.edges = edges
        assert isinstance(expressions, (list_, tuple))
        for e in expressions:
            assert not isinstance(e, (list_, tuple))
        self.expressions = [str(k) for k in expressions]


class AggregatorDescriptorMean(AggregatorDescriptorMulti):
    def __init__(self, name, expressions, short_name="mean", selection=None, edges=False):
        super(AggregatorDescriptorMean, self).__init__(name, expressions, short_name, selection=selection, edges=edges)
        assert len(expressions) == 1

    def add_tasks(self, df, binners, progress):
        progressbar = vaex.utils.progressbars(progress, title=repr(self))
        expression = expression_sum = expression = df[str(self.expressions[0])]

        sum_agg = sum(expression_sum, selection=self.selection, edges=self.edges)
        count_agg = count(expression, selection=self.selection, edges=self.edges)

        task_sum = sum_agg.add_tasks(df, binners, progress=progressbar)[0][0]
        task_count = count_agg.add_tasks(df, binners, progress=progressbar)[0][0]
        self.dtype_in = sum_agg.dtype_in
        self.dtype_out = sum_agg.dtype_out

        @vaex.delayed
        def finish(sum, count):
            sum = np.array(sum)
            dtype = sum.dtype
            sum_kind = sum.dtype.kind
            if sum_kind == 'M':
                sum = sum.view('uint64')
                count = count.view('uint64')
            with np.errstate(divide='ignore', invalid='ignore'):
                mean = sum / count
            if dtype.kind != mean.dtype.kind and sum_kind == "M":
                # TODO: not sure why view does not work
                mean = mean.astype(dtype)
            return mean

        return [task_sum, task_count], finish(task_sum, task_count)


class AggregatorDescriptorVar(AggregatorDescriptorMulti):
    def __init__(self, name, expression, short_name="var", ddof=0, selection=None, edges=False):
        super(AggregatorDescriptorVar, self).__init__(name, expression, short_name, selection=selection, edges=edges)
        self.ddof = ddof

    def add_tasks(self, df, binners, progress):
        progressbar = vaex.utils.progressbars(progress, title=repr(self))
        expression_sum = expression = df[str(self.expressions[0])]
        expression = expression_sum = expression.astype('float64')
        sum_moment = _sum_moment(str(expression_sum), 2, selection=self.selection, edges=self.edges)
        sum_ = sum(str(expression_sum), selection=self.selection, edges=self.edges)
        count_ = count(str(expression), selection=self.selection, edges=self.edges)

        task_sum_moment = sum_moment.add_tasks(df, binners, progress=progressbar)[0][0]
        task_sum = sum_.add_tasks(df, binners, progress=progressbar)[0][0]
        task_count = count_.add_tasks(df, binners, progress=progressbar)[0][0]
        self.dtype_in = sum_.dtype_in
        self.dtype_out = sum_.dtype_out
        @vaex.delayed
        def finish(sum_moment, sum, count):
            sum = np.array(sum)
            dtype = sum.dtype
            if sum.dtype.kind == 'M':
                sum = sum.view('uint64')
                sum_moment = sum_moment.view('uint64')
                count = count.view('uint64')
            with np.errstate(divide='ignore', invalid='ignore'):
                mean = sum / count
                raw_moments2 = sum_moment/count
                variance = (raw_moments2 - mean**2) #* count/(count-self.ddof)
            if dtype.kind != mean.dtype.kind:
                # TODO: not sure why view does not work
                variance = variance.astype(dtype)
            return self.finish(variance)
        return [task_sum_moment, task_sum, task_count], finish(task_sum_moment, task_sum, task_count)


class AggregatorDescriptorSkew(AggregatorDescriptorMulti):
    def __init__(self, name, expression, short_name='skew', selection=None, edges=False):
        super(AggregatorDescriptorSkew, self).__init__(name, [expression], short_name, selection=selection, edges=edges)

    def add_tasks(self, df, binners, progress):
        progressbar = vaex.utils.progressbars(progress, title=repr(self))
        expression = expression_sum = expression = df[str(self.expressions[0])]
        expression = expression_sum = expression.astype('float64')

        sum_moment1 = _sum_moment(str(expression_sum), 1, selection=self.selection, edges=self.edges)
        sum_moment2 = _sum_moment(str(expression_sum), 2, selection=self.selection, edges=self.edges)
        sum_moment3 = _sum_moment(str(expression_sum), 3, selection=self.selection, edges=self.edges)
        count_ = count(str(expression), selection=self.selection, edges=self.edges)

        task_sum_moment1 = sum_moment1.add_tasks(df, binners, progress=progressbar)[0][0]
        task_sum_moment2 = sum_moment2.add_tasks(df, binners, progress=progressbar)[0][0]
        task_sum_moment3 = sum_moment3.add_tasks(df, binners, progress=progressbar)[0][0]
        task_count = count_.add_tasks(df, binners, progress=progressbar)[0][0]

        @vaex.delayed
        def finish(sum_moment1, sum_moment2, sum_moment3, count):
            with np.errstate(divide='ignore', invalid='ignore'):
                m1 = sum_moment1 / count
                m2 = sum_moment2 / count
                m3 = sum_moment3 / count
                skew = (m3 - 3*m1*m2 + 2*m1**3) / (m2 - m1**2)**(3/2)
            return self.finish(skew)
        return [task_sum_moment1, task_sum_moment2, task_sum_moment3, task_count], finish(task_sum_moment1, task_sum_moment2, task_sum_moment3, task_count)


class AggregatorDescriptorKurtosis(AggregatorDescriptorMulti):
    def __init__(self, name, expression, short_name='kurtosis', selection=None, edges=False):
        super(AggregatorDescriptorKurtosis, self).__init__(name, [expression], short_name, selection=selection, edges=edges)

    def add_tasks(self, df, binners, progress):
        progressbar = vaex.utils.progressbars(progress, title=repr(self))
        expression = expression_sum = expression = df[str(self.expressions[0])]
        expression = expression_sum = expression.astype('float64')

        sum_moment1 = _sum_moment(str(expression_sum), 1, selection=self.selection, edges=self.edges)
        sum_moment2 = _sum_moment(str(expression_sum), 2, selection=self.selection, edges=self.edges)
        sum_moment3 = _sum_moment(str(expression_sum), 3, selection=self.selection, edges=self.edges)
        sum_moment4 = _sum_moment(str(expression_sum), 4, selection=self.selection, edges=self.edges)
        count_ = count(str(expression), selection=self.selection, edges=self.edges)

        task_sum_moment1 = sum_moment1.add_tasks(df, binners, progress=progressbar)[0][0]
        task_sum_moment2 = sum_moment2.add_tasks(df, binners, progress=progressbar)[0][0]
        task_sum_moment3 = sum_moment3.add_tasks(df, binners, progress=progressbar)[0][0]
        task_sum_moment4 = sum_moment4.add_tasks(df, binners, progress=progressbar)[0][0]
        task_count = count_.add_tasks(df, binners, progress=progressbar)[0][0]

        @vaex.delayed
        def finish(sum_moment1, sum_moment2, sum_moment3, sum_moment4, count):
            with np.errstate(divide='ignore', invalid='ignore'):
                m1 = sum_moment1 / count
                m2 = sum_moment2 / count
                m3 = sum_moment3 / count
                m4 = sum_moment4 / count
                kurtosis = (m4 - 4*m1*m3 + 6*m1**2*m2 - 3*m1**4) / (m2 - m1**2)**2 -3.0
            return self.finish(kurtosis)
        return [task_sum_moment1, task_sum_moment2, task_sum_moment3, task_sum_moment4, task_count], finish(task_sum_moment1, task_sum_moment2, task_sum_moment3, task_sum_moment4, task_count)


class AggregatorDescriptorStd(AggregatorDescriptorVar):
    def finish(self, value):
        return value**0.5

@register
def count(expression='*', selection=None, edges=False):
    '''Creates a count aggregation'''
    return AggregatorDescriptorBasic('AggCount', [expression], 'count', selection=selection, edges=edges)

@register
def sum(expression, selection=None, edges=False):
    '''Creates a sum aggregation'''
    return AggregatorDescriptorBasic('AggSum', [expression], 'sum', selection=selection, edges=edges)

@register
def mean(expression, selection=None, edges=False):
    '''Creates a mean aggregation'''
    return AggregatorDescriptorMean('mean', [expression], 'mean', selection=selection, edges=edges)

@register
def min(expression, selection=None, edges=False):
    '''Creates a min aggregation'''
    return AggregatorDescriptorBasic('AggMin', [expression], 'min', selection=selection, edges=edges)

@register
def _sum_moment(expression, moment, selection=None, edges=False):
    '''Creates a sum of moment aggregator'''
    return AggregatorDescriptorBasic('AggSumMoment', [expression], '_sum_moment', agg_args=[moment], selection=selection, edges=edges)

@register
def max(expression, selection=None, edges=False):
    '''Creates a max aggregation'''
    return AggregatorDescriptorBasic('AggMax', [expression], 'max', selection=selection, edges=edges)

@register
def first(expression, order_expression=None, selection=None, edges=False):
    '''Creates a first aggregation.

    :param expression: {expression_one}.
    :param order_expression:  Order the values in the bins by this expression.
    :param selection: {selection1}
    :param edges: {edges}
    '''
    return AggregatorDescriptorBasic('AggFirst', [expression, order_expression] if order_expression is not None else [expression], 'first', multi_args=True, selection=selection, edges=edges, agg_args=[False])

@register
@docsubst
def last(expression, order_expression=None, selection=None, edges=False):
    '''Creates a first aggregation.

    :param expression: {expression_one}.
    :param order_expression:  Order the values in the bins by this expression.
    :param selection: {selection1}
    :param edges: {edges}
    '''
    return AggregatorDescriptorBasic('AggFirst', [expression, order_expression] if order_expression is not None else [expression], 'last', multi_args=True, selection=selection, edges=edges, agg_args=[True])

@register
def std(expression, ddof=0, selection=None, edges=False):
    '''Creates a standard deviation aggregation'''
    return AggregatorDescriptorStd('std', [expression], 'std', ddof=ddof, selection=selection, edges=edges)

@register
def var(expression, ddof=0, selection=None, edges=False):
    '''Creates a variance aggregation'''
    return AggregatorDescriptorVar('var', [expression], 'var', ddof=ddof, selection=selection, edges=edges)

@register
def skew(expression, selection=None, edges=False):
    '''Create a skew aggregation.'''
    return AggregatorDescriptorSkew('skew', expression, 'skew', selection=selection, edges=edges)

@register
def kurtosis(expression, selection=None, edges=False):
    '''Create a kurtosis aggregation.'''
    return AggregatorDescriptorKurtosis('kurtosis', expression, 'kurtosis', selection=selection, edges=edges)

@register
@docsubst
def nunique(expression, dropna=False, dropnan=False, dropmissing=False, selection=None, edges=False):
    """Aggregator that calculates the number of unique items per bin.

    :param expression: {expression_one}
    :param dropmissing: {dropmissing}
    :param dropnan: {dropnan}
    :param dropna: {dropna}
    :param selection: {selection1}
    """
    if dropna:
        dropnan = True
        dropmissing = True
    return AggregatorDescriptorNUnique('AggNUnique', [expression], 'nunique', dropmissing, dropnan, selection=selection, edges=edges)

@docsubst
def any(expression=None, selection=None):
    '''Aggregator that returns True when any of the values in the group are True, or when there is any data in the group that is valid (i.e. not missing values or np.nan).
    The aggregator returns False if there is no data in the group when the selection argument is used.

    :param expression: {expression_one}
    :param selection: {selection1}
    '''
    if expression is None and selection is None:
        return count(selection=selection) > -1  # trivial
    else:
        if expression is None:
            return count(selection=selection) > 0
        else:
            return sum(expression, selection=selection) > 0

@docsubst
def all(expression=None, selection=None):
    '''Aggregator that returns True when all of the values in the group are True,
    or when all of the data in the group is valid (i.e. not missing values or np.nan).
    The aggregator returns False if there is no data in the group when the selection argument is used.

    :param expression: {expression_one}
    :param selection: {selection1}
    '''
    if expression is None and selection is None:
        return count(selection=selection) > -1  # trivial
    else:
        if expression is None:
            # counting how often the selection is true == counting how many rows
            return sum(selection) == count(selection)
        else:
            if selection is None:
                # counting how often true == counting how much data there is
                return sum(expression) == count(expression)
            else:
                # since we cannot mix different selections:
                return sum(f'astype({expression}, "bool") & astype({selection}, "bool")') == count(expression)


@register
@docsubst
class list(AggregatorDescriptorBasic):
    '''Aggregator that returns a list of values belonging to the specified expression.

    :param expression: {expression_one}
    :param selection: {selection1}
    :param dropmissing: {dropmissing}
    :param dropnan: {dropnan}
    :param dropna: {dropna}
    :param edges: {edges}
    '''
    def __init__(self, expression, selection=None, dropna=False, dropnan=False, dropmissing=False, edges=False):
        if dropna:
            dropnan = True
            dropmissing = True
        super(list, self).__init__("AggList", [expression], "list", selection=selection, edges=edges, agg_args=[dropnan, dropmissing])

    def _prepare_types(self, df):
        super()._prepare_types(df)
        self.dtype_out = vaex.dtype(pa.large_list(self.dtype_out.arrow))


# TODO: we could maybe refactor this to support more struct aggregators
@register
class describe(AggregatorDescriptor):
    def __init__(self, expression):
        self.expression = expression
        self.expressions = [self.expression]
        self.short_name = "describe"
        self.edges = True

    def __repr__(self):
        return f"describe({self.expression!r})"

    def add_tasks(self, df, binners, progress):
        expression: Expression = df[str(self.expression)]
        col = expression._label
        if expression.data_type() != "string":
            aggs = {
                f"count": vaex.agg.count(expression, edges=self.edges),
                f"count_na": vaex.agg.count(edges=self.edges) - vaex.agg.count(expression, edges=self.edges),
                f"mean": vaex.agg.mean(expression, edges=self.edges),
                f"std": vaex.agg.std(expression, edges=self.edges),
                f"min": vaex.agg.min(expression, edges=self.edges),
                f"max": vaex.agg.max(expression, edges=self.edges),
            }
        else:
            aggs = {f"count": vaex.agg.count(expression, edges=self.edges), f"count_na": vaex.agg.count(edges=self.edges) - vaex.agg.count(expression, edges=self.edges)}

        progressbar = vaex.utils.progressbars(progress, title=repr(self))
        tasks = []
        results = []
        names = []
        for name, agg in aggs.items():
            tasks1, result = agg.add_tasks(df, binners, progress=progressbar)
            tasks.extend(tasks1)
            results.append(result)
            names.append(name)

        @vaex.delayed
        def finish(*values):
            return self.finish(values, names)

        return tasks, finish(*results)

    def finish(self, values, names):
        if len(values):
            if vaex.array_types.is_scalar(values[0]):
                return pa.StructArray.from_arrays(arrays=[[k] for k in values], names=names)
        return pa.StructArray.from_arrays(arrays=values, names=names)


# @register
# def covar(x, y):
#     '''Creates a standard deviation aggregation'''
#     return _Statistic('covar', x, y)

# @register
# def correlation(x, y):
#     '''Creates a standard deviation aggregation'''
#     return _Statistic('correlation', x, y)



@dask.base.normalize_token.register(AggregatorDescriptor)
def normalize(agg):
    return agg.__class__.__name__, repr(agg)



# -*- coding: utf-8 -*-
from __future__ import division
import logging
import collections
import ast
import _ast
import string
import numpy as np
import math
import sys
import six
import copy
import difflib


if hasattr(_ast, 'Num'):
    ast_Num = _ast.Num
    ast_Str = _ast.Str
else:  # Python3.8
    ast_Num = _ast.Constant
    ast_Str = _ast.Constant

if hasattr(_ast, 'NameConstant'):
    ast_Constant = _ast.NameConstant
else:
    ast_Constant = _ast.Constant


logger = logging.getLogger("expr")
logger.setLevel(logging.ERROR)


valid_binary_operators = [_ast.Add, _ast.Sub, _ast.Mult, ast.MatMult, _ast.Pow,
                          _ast.Div, _ast.FloorDiv, _ast.BitAnd, _ast.BitOr, _ast.BitXor, _ast.Mod,
                          _ast.RShift, _ast.LShift
                          ]
valid_compare_operators = [_ast.Lt, _ast.LtE,
                           _ast.Gt, _ast.GtE, _ast.Eq, _ast.NotEq, _ast.IsNot, _ast.Is, _ast.In]
valid_unary_operators = [_ast.USub, _ast.UAdd, _ast.Invert]
valid_id_characters = string.ascii_letters + string.digits + "_"
valid_functions = "sin cos".split()

opmap = {
    _ast.Add: '+',
    _ast.Sub: '-',
    _ast.Mult: '*',
    _ast.Pow: '**',
    _ast.Div: '/',
    _ast.FloorDiv: '//',
    _ast.BitAnd: '&',
    _ast.BitOr: '|',
    _ast.BitXor: '^',
    _ast.Mod: '%',
}


def math_parse(expression, macros=[]):
    # TODO: validate macros?
    node = ast.parse(expression)
    if len(node.body) != 1:
        raise ValueError("expected one expression, got %r" % len(node.body))
    expr = node.body[0]
    if not isinstance(expr, _ast.Expr):
        raise ValueError("expected an expression got a %r" % type(node.body))

    validate_expression(expr.value)
    return MathExpression(expression, macros)


last_func = None


def validate_expression(expr, variable_set, function_set=[], names=None):
    global last_func
    names = names if names is not None else []
    if isinstance(expr, six.string_types):
        node = ast.parse(expr)
        if len(node.body) != 1:
            raise ValueError("expected one expression, got %r" %
                             len(node.body))
        first_expr = node.body[0]
        if not isinstance(first_expr, _ast.Expr):
            raise ValueError("expected an expression got a %r" %
                             type(node.body))
        validate_expression(first_expr.value, variable_set,
                            function_set, names)
    elif isinstance(expr, _ast.BinOp):
        if expr.op.__class__ in valid_binary_operators:
            validate_expression(expr.right, variable_set, function_set, names)
            validate_expression(expr.left, variable_set, function_set, names)
        else:
            raise ValueError("Binary operator not allowed: %r" % expr.op)
    elif isinstance(expr, _ast.UnaryOp):
        if expr.op.__class__ in valid_unary_operators:
            validate_expression(expr.operand, variable_set,
                                function_set, names)
        else:
            raise ValueError("Unary operator not allowed: %r" % expr.op)
    elif isinstance(expr, _ast.Name):
        if expr.id not in variable_set:
            matches = difflib.get_close_matches(expr.id, list(variable_set))
            msg = "Column or variable %r does not exist." % expr.id
            if matches:
                msg += ' Did you mean: ' + " or ".join(map(repr, matches))

            raise NameError(msg)
        names.append(expr.id)
    elif isinstance(expr, ast_Num):
        pass  # numbers are fine
    elif isinstance(expr, ast_Str):
        pass  # as well as strings
    elif isinstance(expr, _ast.Call):
        validate_func(expr.func, function_set)
        last_func = expr
        for arg in expr.args:
            validate_expression(arg, variable_set, function_set, names)
        for arg in expr.keywords:
            validate_expression(arg, variable_set, function_set, names)
    elif isinstance(expr, _ast.Compare):
        validate_expression(expr.left, variable_set, function_set, names)
        for op in expr.ops:
            if op.__class__ not in valid_compare_operators:
                raise ValueError("Compare operator not allowed: %r" % op)
        for comparator in expr.comparators:
            validate_expression(comparator, variable_set, function_set, names)
    elif isinstance(expr, _ast.keyword):
        validate_expression(expr.value, variable_set, function_set, names)
    elif isinstance(expr, ast_Constant):
        pass  # like True and False
    elif isinstance(expr, _ast.List):
        for el in expr.elts:
            validate_expression(el, variable_set, function_set, names)
    elif isinstance(expr, _ast.Dict):
        for key in expr.keys:
            validate_expression(key, variable_set, function_set, names)
        for value in expr.values:
            validate_expression(value, variable_set, function_set, names)
    elif isinstance(expr, _ast.Subscript):
        validate_expression(expr.value, variable_set, function_set, names)
        if isinstance(expr.slice.value, ast_Num):
            pass  # numbers are fine
        elif isinstance(expr.slice.value, str) or isinstance(expr.slice.value, _ast.Str):
            pass  # and strings (from py3.9, value is str)
        else:
            raise ValueError(
                "Only subscript/slices with numbers allowed, not: %r" % expr.slice.value)
    else:
        last_func = expr
        raise ValueError("Unknown expression type: %r" % type(expr))


class Validator(ast.NodeVisitor):

    def generic_visit(self, node):
        raise ValueError('unexpected node: {}', ast.dump(node))

    def visit_BinOp(self, expr):
        if expr.op.__class__ in valid_binary_operators:
            validate_expression(expr.right, variable_set, function_set, names)
            validate_expression(expr.left, variable_set, function_set, names)
        else:
            raise ValueError("Binary operator not allowed: %r" % expr.op)


def mul(left, right):
    return ast.BinOp(left=left, right=right, op=ast.Mult())


def div(left, right):
    return ast.BinOp(left=left, right=right, op=ast.Div())


def add(left, right):
    return ast.BinOp(left=left, right=right, op=ast.Add())


def sub(left, right):
    return ast.BinOp(left=left, right=right, op=ast.Sub())


def pow(left, right):
    return ast.BinOp(left=left, right=right, op=ast.Pow())


def sqr(node):
    return ast.BinOp(left=node, right=num(2), op=ast.Pow())

def sqrt(node):
    return call('sqrt', [node])


def neg(node):
    return ast.UnaryOp(op=ast.USub(), operand=node)


def num(n):
    return ast.Num(n=n)


def call(fname, args):
    return ast.Call(func=ast.Name(id=fname, ctx=ast.Load()), args=args)


def _dlog10(n, args):
    assert len(args) == 1
    assert n == 0
    a = call('log', args=[num(10)])
    return div(num(1), mul(args[0], a))


def _dsqrt(n, args):
    assert n == 0
    assert len(args) == 1
    a = call('log', args=[num(10)])
    return mul(num(1/2), pow(args[0], num(-0.5)))


def _dcos(n, args):
    assert n == 0
    assert len(args) == 1
    return neg(call('sin', args=args))

def _darccos(n, args):
    assert n == 0
    assert len(args) == 1
    a = sqrt(sub(num(1), sqr(args[0])))
    return neg(div(num(1), a))

def _darctan2(n, args):
    # derivative of arctan2(y, x)
    assert (n >= 0) and (n <= 1)
    assert len(args) == 2
    y, x = args
    if n == 1:  # derivative wrt 2nd argument (x)
        return div(neg(y), add(sqr(x), sqr(y)))
    if n == 0:  # derivative wrt 1st argument (y)
        return div(x, add(sqr(x), sqr(y)))

def _dtan(n, args):
    assert n == 0
    assert len(args) == 1
#     a = div(sub(num(1), sqr(args[0])))
    return div(num(1), sqr(call('cos', args=args)))

standard_function_derivatives = {}
standard_function_derivatives['sin'] = 'cos'
standard_function_derivatives['cos'] = _dcos
standard_function_derivatives['tan'] = _dtan
standard_function_derivatives['log10'] = _dlog10
standard_function_derivatives['sqrt'] = _dsqrt
standard_function_derivatives['arctan2'] = _darctan2
standard_function_derivatives['arccos'] = _darccos


class Derivative(ast.NodeTransformer):
    def __init__(self, id, function_derivatives={}):
        self.id = id
        self.function_derivatives = dict(standard_function_derivatives)
        self.function_derivatives.update(function_derivatives)

    def format(self, node):
        # try:
        return ExpressionString().visit(node)
        # return ast.dump(node)

    def visit_Num(self, node):
        return ast.Num(n=0)

    def visit_Name(self, node):
        if node.id == self.id:
            return ast.Num(n=1)
        else:
            return ast.Num(n=0)

    def visit_Call(self, node):
        fname = node.func.id
        df = self.function_derivatives.get(fname)
        if df is None:
            raise ValueError('Derivative of {} is unknown'.format(fname))
        if not callable(df):  # simply a string
            assert len(node.args) == 1
            result = mul(call(df, node.args), self.visit(node.args[0]))
        else:
            terms = [mul(df(i, node.args), self.visit(arg))
                     for i, arg in enumerate(node.args)]
            result = terms[0]
            for term in terms[1:]:
                result = add(result, term)
        return result

    def generic_visit(self, node):
        # it's annoying that the default one modifies in place
        return super(Derivative, self).generic_visit(copy.deepcopy(node))

    def visit_BinOp(self, node):
        solution = None
        if isinstance(node.op, ast.Mult):
            solution = add(mul(self.visit(node.left), node.right),
                           mul(node.left,             self.visit(node.right)))
        if isinstance(node.op, ast.Div):
            # (n*at - t*an) / n2
            n = node.right
            t = node.left
            at = self.visit(t)
            an = self.visit(n)
            solution = div(sub(mul(n, at), mul(t, an)), pow(n, num(2)))
        if isinstance(node.op, ast.Add):
            solution = add(self.visit(node.left), self.visit(node.right))
        if isinstance(node.op, ast.Sub):
            solution = sub(self.visit(node.left), self.visit(node.right))
        if isinstance(node.op, ast.Pow):
            # following https://en.wikipedia.org/wiki/Differentiation_rules
            f = node.left
            df = self.visit(f)
            g = node.right
            dg = self.visit(g)
            # if g is a number, we take a equivalent solution, which gives a nicer result
            if isinstance(g, ast.Num):
                solution = mul(g, mul(df, pow(node.left, num(node.right.n-1))))
            else:
                a = add(mul(df, div(g, f)), mul(dg, call('log', [f])))
                solution = mul(pow(f, g), a)
        if solution is None:
            raise ValueError('Unknown rule for: {}'.format(self.format(node)))
        return solution


class ExpressionString(ast.NodeVisitor):
    def __init__(self, pretty=False):
        self.pretty = pretty
        self.indent = 0
    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.USub):
            if isinstance(node.operand, (ast.Name, ast.Num, _ast.Name)):
                return "-{}".format(self.visit(node.operand))  # prettier
            else:
                return "-({})".format(self.visit(node.operand))
        elif isinstance(node.op, ast.UAdd):
            if isinstance(node.operand, (ast.Name, ast.Num, _ast.Name)):
                return "+{}".format(self.visit(node.operand))  # prettier
            else:
                return "+({})".format(self.visit(node.operand))
        elif isinstance(node.op, ast.Invert):
            if isinstance(node.operand, (ast.Name, ast.Num, _ast.Name)):
                return "~{}".format(self.visit(node.operand))  # prettier
            else:
                return "~({})".format(self.visit(node.operand))
        else:
            raise ValueError('Unary op not supported: {}'.format(node.op))

    def visit_Name(self, node):
        return node.id

    def visit_Num(self, node):
        return repr(node.n)

    def visit_keyword(self, node):
        return "%s=%s" % (node.arg, self.visit(node.value))

    def visit_NameConstant(self, node):
        return repr(node.value)

    def visit_Dict(self, node):
        parts = []
        for key, value in zip(node.keys, node.values):
            key = self.visit(key)
            value = self.visit(value)
            parts.append(f'{key}: {value}')
        return '{' + ', '.join(parts) + '}'

    def visit_Call(self, node):
        args = [self.visit(k) for k in node.args]
        keywords = []
        if hasattr(node, 'keywords'):
            keywords = [self.visit(k) for k in node.keywords]
        return "{}({})".format(node.func.id, ", ".join(args + keywords))

    def visit_Str(self, node):
        return repr(node.s)

    def visit_List(self, node):
        return "[{}]".format(", ".join([self.visit(k) for k in node.elts]))

    def pow(self, left, right):
        return "({left} ** {right})".format(left=left, right=right)

    def visit_BinOp(self, node):
        newline = indent = ""
        if self.pretty:
            indent = "  " * self.indent
            newline = "\n"
        self.indent += 1
        left = "{}{}{}".format(newline, indent, self.visit(node.left))
        right = "{}{}{}".format(newline, indent, self.visit(node.right))
        try:
            if isinstance(node.op, ast.Mult):
                return "({left} * {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.MatMult):
                return "({left} @ {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.Div):
                return "({left} / {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.Mod):
                return "({left} % {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.FloorDiv):
                return "({left} // {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.Add):
                return "({left} + {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.Sub):
                return "({left} - {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.Pow):
                return self.pow(left, right)
            elif isinstance(node.op, ast.BitAnd):
                return "({left} & {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.BitOr):
                return "({left} | {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.BitXor):
                return "({left} ^ {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.RShift):
                return "({left} >> {right})".format(left=left, right=right)
            elif isinstance(node.op, ast.LShift):
                return "({left} << {right})".format(left=left, right=right)
            else:
                raise ValueError(f'Do not know binary op {node.op}')
                # return "do_not_understand_expression"
        finally:
            self.indent -= 1

    op_translate = {ast.Lt: "<", ast.LtE: "<=", ast.Gt: ">", ast.GtE: ">=", ast.Eq: "==", ast.NotEq: "!=",
                    ast.IsNot: "is not", ast.Is: "is", ast.In: "in"}
    def visit_Compare(self, node):
        s = ""
        left = self.visit(node.left)
        for op, comp in zip(node.ops, node.comparators):
            right = self.visit(comp)
            op = ExpressionString.op_translate[op.__class__]
            s = "({left} {op} {right})".format(left=left, op=op, right=right)
            left = right
        return s

    def visit_Subscript(self, node):
        p = self.visit(node.value)
        v = self.visit(node.slice.value)
        return f'{p}[{v}]'

    # required from py3.9, since in visit_Subscript node can be a string
    def visit_str(self, node):
        return repr(node)

class SimplifyExpression(ast.NodeTransformer):

    def visit_UnaryOp(self, node):
        node.operand = self.visit(node.operand)
        if isinstance(node.op, ast.USub):
            if isinstance(node.operand, ast.Num) and node.operand.n == 0:
                node = node.operand
        return node

    def visit_BinOp(self, node):
        node.left = left = self.visit(node.left)
        node.right = right = self.visit(node.right)
        if isinstance(node.op, ast.Mult):
            if isinstance(right, ast.Num) and right.n == 0:
                return num(0)
            elif isinstance(right, ast.Num) and right.n == 1:
                return left
            elif isinstance(left, ast.Num) and left.n == 0:
                return num(0)
            elif isinstance(left, ast.Num) and left.n == 1:
                return right
        if isinstance(node.op, ast.Div):
            if isinstance(left, ast.Num) and left.n == 0:
                return num(0)
        if isinstance(node.op, ast.Add):
            if isinstance(right, ast.Num) and right.n == 0:
                return left
            if isinstance(left, ast.Num) and left.n == 0:
                return right
        if isinstance(node.op, ast.Sub):
            if isinstance(right, ast.Num) and right.n == 0:
                return left
            if isinstance(left, ast.Num) and left.n == 0:
                return neg(right)
        if isinstance(node.op, ast.Pow):
            if isinstance(left, ast.Num) and left.n == 0:
                return num(0)  # not ok with negative powers..
            if isinstance(right, ast.Num) and right.n == 0:
                # TODO: this means a numpy arrays can become a scalar
                return num(1)
            if isinstance(right, ast.Num) and right.n == 1:
                return left
        return node


class Translator(ast.NodeTransformer):
    def __init__(self, translator):
        self.translator = translator

    def visit_Call(self, node):
        # we skip visiting node.id
        node.args = [self.visit(k) for k in node.args]
        if hasattr(node, 'keywords'):
            node.keywords = [self.visit(k) for k in node.keywords]
        return node

    def visit_Name(self, node):
        expr = self.translator(node.id)
        if expr:
            node = parse_expression(expr)
            node = self.visit(node)
        return node


class NameCollector(ast.NodeTransformer):
    def __init__(self):
        self.names = {}

    def visit_Call(self, node):
        # we skip visiting node.id
        self.visit(node.func)
        node.args = [self.visit(k) for k in node.args]
        if hasattr(node, 'keywords'):
            node.keywords = [self.visit(k) for k in node.keywords]
        return node

    def visit_Name(self, node):
        if node.id not in self.names:
            self.names[node.id] = []
        self.names[node.id].append(node)
        return node


class SliceCollector(ast.NodeTransformer):
    def __init__(self):
        self.slices = collections.defaultdict(list)

    def visit_Subscript(self, node):
        # py39
        if node.value.id == 'df' and isinstance(node.slice.value, str):
            self.slices[node.slice.value].append(node)
        if node.value.id == 'df' and isinstance(node.slice.value, ast.Str):
            self.slices[node.slice.value.s].append(node)
        return node

class GraphBuiler(ast.NodeVisitor):
    def __init__(self):
        self.dependencies = []

    def visit_Call(self, node):
        fname = node.func.id
        dependencies = list(self.dependencies)
        self.dependencies = []
        for arg in node.args:
            self.visit(arg)
        graph = [fname, node_to_string(node), self.dependencies]
        dependencies.append(graph)
        self.dependencies = dependencies

    def visit_BinOp(self, node):
        dependencies = list(self.dependencies)
        self.dependencies = []
        self.visit(node.left)
        dep_left = self.dependencies

        self.dependencies = []
        self.visit(node.right)
        dep_right = self.dependencies
        graph = [opmap[type(node.op)], node_to_string(node), dep_left + dep_right]
        dependencies.append(graph)
        self.dependencies = dependencies

    def visit_Name(self, node):
        self.dependencies.append(node.id)


def _graph(expression_string):
    node = parse_expression(expression_string)
    g = GraphBuiler()
    node = g.visit(node)
    return g.dependencies[0]


def simplify(expression_string):
    node = parse_expression(expression_string)
    node = SimplifyExpression().visit(node)
    return node_to_string(node)


def derivative(expression, variable_name, simplify=True):
    if isinstance(expression, str):
        node = parse_expression(expression)
    else:
        node = expression
    node = Derivative(variable_name).visit(node)
    if simplify:
        node = SimplifyExpression().visit(node)
    return node_to_string(node)


def translate(expression, translator):
    if isinstance(expression, str):
        node = parse_expression(expression)
    else:
        node = expression
    node = Translator(translator).visit(node)
    return node_to_string(node)


def names(expression):
    if isinstance(expression, str):
        node = parse_expression(expression)
    else:
        node = expression
    nc = NameCollector()
    nc.visit(node)
    return nc.names


def slices(expression):
    if isinstance(expression, str):
        node = parse_expression(expression)
    else:
        node = expression
    nc = SliceCollector()
    nc.visit(node)
    return nc.slices


def parse_expression(expression_string):
    expr = ast.parse(expression_string).body[0]
    assert isinstance(expr, ast.Expr), f"not an expression {str(expr)}"
    return expr.value


def node_to_string(node, pretty=False):
    return ExpressionString(pretty=pretty).visit(node)


def validate_func(name, function_set):
    if name.id not in function_set:
        raise NameError("function %r is not defined" % name.id)



from vaex.utils import _issequence


class Rolling:
    '''Provides rolling window calculations'''
    def __init__(self, df, window, trim, edge, fill_value=None, columns=None):
        self.df = df
        self.window = window
        self.trim = trim
        self.edge = edge
        self.fill_value = fill_value
        self.columns = columns

    def array(self):
        '''Creates an array representing each window'''
        if self.edge == "right":
            return self.df.shift((-self.window, 0), column=self.columns, trim=self.trim, fill_value=self.fill_value)
        elif self.edge == "left":
            return self.df.shift((0, self.window), column=self.columns, trim=self.trim, fill_value=self.fill_value)
        else:
            raise ValueError(f'edge can be "right", "left" or "center", not {self.edge}')

    def sum(self):
        '''Sum all values in the window'''
        df = self.array()
        for column in self.columns:
            df[column] = df[column].sum(axis=1)
        return df

    # def apply(self, f):
    #     df = self.array()
    #     for column in self.columns:
    #         df[column] = df[column].apply(f)
    #     return df



__author__ = 'maartenbreddels'
import unittest
import os
import tempfile
import logging
import shutil

import numpy as np
import PIL.Image
import PIL.ImageChops
import matplotlib.pyplot as plt

import vaex as vx
import vaex.utils

try:
	raw_input = input
except:
	pass # py2/3 fix

base_path = os.path.dirname(__file__)
def get_comparison_image(name):
	osname = vaex.utils.osname
	return os.path.join(base_path, "images", name+"_" + osname + ".png")

overwrite_images = False
class check_output(object):
	def __init__(self, name):
		self.name = name
		self.fn = get_comparison_image(name)


	def __enter__(self):
		plt.figure()

	def __exit__(self, exc_type, exc_val, exc_tb):
		if exc_type is not None:
			plt.close()
			return
		fn = tempfile.mktemp(".png")
		plt.savefig(fn)
		if not os.path.exists(self.fn):
			print("comparison image did not exist, copying to %s" % self.fn)
			shutil.copy(fn, self.fn)

		image1 = PIL.Image.open(self.fn)
		image2 = PIL.Image.open(fn)
		diff = PIL.ImageChops.difference(image1, image2)
		extrema = diff.getextrema()
		for i, (vmin, vmax) in enumerate(extrema):
			msg = "difference found between {im1} and {im2} in band {band}\n $ cp {im2} {im1}".format(im1=self.fn, im2=fn,
																									  band=i)
			if vmin != vmax and overwrite_images:
				image1.show()
				image2.show()
				done = False
				while not done:
					answer = raw_input("is the new image ok? [y/N]").lower().strip()
					if answer == "n":
						self.assertEqual(vmin, 0, msg)
						return
					if answer == "y":
						shutil.copy(fn, self.fn)
						return
			assert vmin == 0, msg
			assert vmax == 0, msg
		plt.close()


class TestPlot(unittest.TestCase):
	def setUp(self):
		self.dataset = vx.example()

	def tearDown(self):
		if vx.utils.osname != "osx":
			self.dataset.close()

	def test_single(self):
		with check_output("single_xy"):
			self.dataset.plot("x", "y", title="face on")

	def test_single_nan(self):
		with check_output("single_xy_no_nan"):
			self.dataset.plot("x", "y", f="log")
		cm = plt.cm.inferno
		cm.set_bad("orange")
		with check_output("single_xy_nan"):
			self.dataset.plot("x", "y", f="log", colormap=cm)

	def test_multiplot(self):
		with check_output("multiplot_xy"):
			self.dataset.plot([["x", "y"], ["x", "z"]], title="Face on and edge on", figsize=(10, 4));

	def test_multistat(self):
		with check_output("multistat"):
			self.dataset.plot("x", "y", what=["count(*)", "mean(vx)", "correlation(vy, vz)"], title="Different statistics", figsize=(10,5));

	def test_multiplot_multiwhat(self):
		with check_output("multiplot_multiwhat"):
			self.dataset.plot([["x", "y"], ["x", "z"], ["y", "z"]],
				  what=["count(*)", "mean(vx)", "correlation(vx, vy)", "correlation(vx, vz)"],
				  title="Different statistics and plots", figsize=(14, 12));

	def test_multistat_multiwhat_swapped(self):
		with check_output("multistat_multiwhat_swapped"):
			self.dataset.plot([["x", "y"], ["x", "z"], ["y", "z"]],
				what=["count(*)", "mean(vx)", "correlation(vx, vy)", "correlation(vx, vz)"],
				visual=dict(row="what", column="subspace"),
				title="Different statistics and plots", figsize=(14,12));

	def test_slice(self):
		with check_output("slice"):
			self.dataset.plot("Lz", "E", z="FeH:-3,-1,10", visual=dict(row="z"), figsize=(12,8), f="log", wrap_columns=3);

	def test_plot1d(self):
		with check_output("plot1d"):
			self.dataset.plot1d("Lz");

	def test_scatter(self):
		self.dataset.set_active_fraction(0.01)
		with check_output("scatter"):
			self.dataset.scatter("Lz", "E", length_check=False);
		with check_output("scatter_xerr"):
			self.dataset.scatter("Lz", "Lz", xerr="abs(Lz*0.1)", length_check=False);
		with check_output("scatter_xerr_yerr"):
			self.dataset.scatter("Lz", "Lz", xerr="abs(Lz*0.1)", yerr="abs(Lz*0.4)", length_check=False);
		with check_output("scatter_xerr_yerr_asym"):
			self.dataset.scatter("Lz", "Lz", xerr=["abs(Lz*0.1)", "abs(Lz*0.2)"], yerr=["abs(Lz*0.2)", "abs(Lz)"], length_check=False);


	def test_healpix(self):
		self.dataset.add_virtual_columns_cartesian_to_spherical()
		self.dataset.add_column_healpix(longitude="l", latitude="b")
		with check_output("plot_healpix"):
			self.dataset.healpix_plot("healpix");

if __name__ == '__main__':
    unittest.main()



__author__ = 'maartenbreddels'
import unittest

def load_tests(loader, tests, pattern):
	import vaex.test.dataset
	import vaex.test.ui
	import vaex.test.misc
	import vaex.test.plot
	import vaex.test.cmodule
	print(loader, tests, pattern)
	for module in [vaex.test.dataset, vaex.test.ui, vaex.test.misc, vaex.test.cmodule, vaex.test.plot]:
		tests.addTests(unittest.TestLoader().loadTestsFromModule(module))
	return tests





from abc import  abstractmethod, abstractproperty
import difflib
import os
from pathlib import Path
import collections.abc
import logging
import uuid
from urllib.parse import urlparse
from typing import Set, List
import threading

import numpy as np
from frozendict import frozendict
import pyarrow as pa

import vaex
import vaex.execution
import vaex.settings
import vaex.utils
from vaex.array_types import data_type
from .column import Column, ColumnIndexed, supported_column_types
from . import array_types
from vaex import encoding
try:
    from sys import version_info
    if version_info[:2] >= (3, 10):
        from importlib.metadata import entry_points
    else:
        from importlib_metadata import entry_points
except ImportError:
    import pkg_resources
    entry_points = pkg_resources.iter_entry_points

logger = logging.getLogger('vaex.dataset')

opener_classes = []
HASH_VERSION = "1"
HASH_VERSION_KEY = "version"
chunk_size_default = vaex.settings.main.chunk.size or 1024**2

_dataset_types = {}
lock = threading.Lock()


def register(cls, name=None):
    name = name or getattr(cls, 'snake_name') or cls.__name__
    _dataset_types[name] = cls
    return cls

@encoding.register('dataset')
class dataset_encoding:
    @staticmethod
    def encode(encoding, dataset):
        return dataset.encode(encoding)

    @staticmethod
    def decode(encoding, dataset_spec):
        dataset_spec = dataset_spec.copy()
        type = dataset_spec.pop('dataset_type')
        cls = _dataset_types[type]
        return cls.decode(encoding, dataset_spec)


def open(path, fs_options={}, fs=None, *args, **kwargs):
    failures = []
    with lock:  # since we cache, make this thread save
        if not opener_classes:
            for entry in entry_points(group='vaex.dataset.opener'):
                logger.debug('trying opener: ' + entry.name)
                try:
                    opener = entry.load()
                    opener_classes.append(opener)
                except Exception as e:
                    logger.exception('issue loading ' + entry.name)
                    failures.append((e, entry))

    # first the quick path
    for opener in opener_classes:
        if opener.quick_test(path, fs_options=fs_options, fs=fs):
            if opener.can_open(path, fs_options=fs_options, fs=fs, *args, **kwargs):
                return opener.open(path, fs_options=fs_options, fs=fs, *args, **kwargs)

    # otherwise try all openers
    for opener in opener_classes:
        try:
            if opener.can_open(path, fs_options=fs_options, fs=fs, *args, **kwargs):
                return opener.open(path, fs_options=fs_options, fs=fs, *args, **kwargs)
        except Exception as e:
            failures.append((e, opener))

    failures = "\n".join([f'\n-----{who}-----\n:' + vaex.utils.format_exception_trace(e) for e, who in failures])
    if failures:
        raise IOError(f'Cannot open {path}, failures: {failures}.')
    else:
        raise IOError(f'Cannot open {path} nobody knows how to read it.')


def _to_bytes(ar):
    try:
        return ar.view(np.uint8)
    except ValueError:
        return ar.copy().view(np.uint8)

def hash_combine(*hashes):
    hasher = vaex.utils.create_hasher(large_data=False)
    for hash in hashes:
        hasher.update(hash.encode())
    return hasher.hexdigest()


def hash_slice(hash, start, end):
    hasher = vaex.utils.create_hasher(hash.encode(), large_data=False)
    slice = np.array([start, end], dtype=np.int64)
    hasher.update(_to_bytes(slice))
    return hasher.hexdigest()


def hash_array_data(ar):
    # this function should stay consistent with all future versions
    # since this is the expensive part of the hashing
    if isinstance(ar, np.ndarray):
        ar = ar.ravel()
        if ar.dtype == np.object_:
            return {"type": "numpy", "data": str(uuid.uuid4()), "mask": None}
        if np.ma.isMaskedArray(ar):
            data_byte_ar = _to_bytes(ar.data)
            hasher = vaex.utils.create_hasher(data_byte_ar, large_data=True)
            hash_data = {"type": "numpy", "data": hasher.hexdigest(), "mask": None}
            if ar.mask is not True and ar.mask is not False and ar.mask is not np.True_ and ar.mask is not np.False_:
                mask_byte_ar = _to_bytes(ar.mask)
                hasher = vaex.utils.create_hasher(mask_byte_ar, large_data=True)
                hash_data["mask"] = hasher.hexdigest()
            return hash_data
        else:
            try:
                byte_ar = _to_bytes(ar)
            except ValueError:
                byte_ar = ar.copy().view(np.uint8)
            hasher = vaex.utils.create_hasher(byte_ar, large_data=True)
            hash_data = {"type": "numpy", "data": hasher.hexdigest(), "mask": None}
    elif isinstance(ar, (pa.Array, pa.ChunkedArray)):
        hasher = vaex.utils.create_hasher(large_data=True)
        buffer_hashes = []
        hash_data = {"type": "arrow", "buffers": buffer_hashes}
        if isinstance(ar, pa.ChunkedArray):
            chunks = ar.chunks
        else:
            chunks = [ar]
        for chunk in chunks:
            for buffer in chunk.buffers():
                if buffer is not None:
                    hasher.update(memoryview(buffer))
                    buffer_hashes.append(hasher.hexdigest())
                else:
                    buffer_hashes.append(None)
    elif isinstance(ar, vaex.column.Column):
        hash_data = {"type": "column", "fingerprint": ar.fingerprint()}
    else:
        raise TypeError
    return hash_data


def hash_array(ar, hash_info=None, return_info=False):
    # this function can change over time, as it builds on top of the expensive part
    # (hash_array_data), so we can cheaply calculate new hashes if we pass on hash_info
    if hash_info is None:
        hash_info = hash_array_data(ar)
    if hash_info.get(HASH_VERSION_KEY) == HASH_VERSION:  # TODO: semver check?
        return hash_info['hash'], hash_info
    if isinstance(ar, np.ndarray):
        if ar.dtype == np.object_:
            return hash_info['data']  # uuid, so always unique
        if np.ma.isMaskedArray(ar):
            if not (hash_info['type'] == 'numpy' and hash_info['data'] and hash_info['mask']):
                hash_info = hash_array_data(ar)
        else:
            if not (hash_info['type'] == 'numpy' and hash_info['data']):
                hash_info = hash_array_data(ar)
        keys = [HASH_VERSION, hash_info['type'], hash_info['data']]
        if hash_info['mask']:
            keys.append(hash_info['mask'])
    elif isinstance(ar, vaex.array_types.supported_arrow_array_types):
        if not (hash_info['type'] == 'arrow' and hash_info['buffers']):
            hash_info = hash_array_data(ar)
        keys = [HASH_VERSION]
        keys.extend(["NO_BUFFER" if not b else b for b in hash_info['buffers']])
    elif isinstance(ar, vaex.column.Column):
        if not (hash_info['type'] == 'column'):
            hash_info = hash_array_data(ar)
        keys = [HASH_VERSION]
        keys.append(hash_info['fingerprint'])
    hasher = vaex.utils.create_hasher(large_data=False)  # small amounts of data
    for key in keys:
        hasher.update(key.encode('ascii'))
    hash = hasher.hexdigest()
    if return_info:
        hash_info['hash'] = hash
        hash_info[HASH_VERSION_KEY] = HASH_VERSION
        return hash, hash_info
    else:
        return hash


def to_supported_array(ar):
    if not isinstance(ar, supported_column_types):
        ar = np.asanyarray(ar)

    if isinstance(ar, np.ndarray) and ar.dtype.kind == 'U':
        ar = vaex.column.ColumnArrowLazyCast(ar, pa.string())
    elif isinstance(ar, np.ndarray) and ar.dtype.kind == 'O':
        ar_data = ar
        if np.ma.isMaskedArray(ar):
            ar_data = ar.data

        try:
            # "k != k" is a way to detect NaN's and NaT's
            types = list({type(k) for k in ar_data if k is not None and k == k})
        except ValueError:
            # If there is an array value in the column, Numpy throws a ValueError
            # "The truth value of an array with more than one element is ambiguous".
            # We don't handle this by default as it is a bit slower.
            def is_missing(k):
                if k is None:
                    return True
                try:
                    # a way to detect NaN's and NaT
                    return not (k == k)
                except ValueError:
                    # if a value is an array, this will fail, and it is a non-missing
                    return False
            types = list({type(k) for k in ar_data if k is not is_missing(k)})

        if len(types) == 1 and issubclass(types[0], str):
            # TODO: how do we know it should not be large_string?
            # self._dtypes_override[valid_name] = pa.string()
            ar = vaex.column.ColumnArrowLazyCast(ar, pa.string())
        if len(types) == 0:  # can only be if all nan right?
            ar = ar.astype(np.float64)
    return ar


def _concat_chunk_list(list_of_chunks):
    dict_of_list_of_arrays = collections.defaultdict(list)
    for chunks in list_of_chunks:
        for name, array in chunks.items():
            if isinstance(array, pa.ChunkedArray):
                dict_of_list_of_arrays[name].extend(array.chunks)
            else:
                dict_of_list_of_arrays[name].append(array)
    chunks = {name: vaex.array_types.concat(arrays) for name, arrays in dict_of_list_of_arrays.items()}
    return chunks


def _slice_of_chunks(chunks_ready_list, chunk_size):
    current_row_count = 0
    chunks_current_list = []
    while current_row_count < chunk_size and chunks_ready_list:
        chunks_current = chunks_ready_list.pop(0)
        chunk = list(chunks_current.values())[0]
        # chunks too large, split, and put back a part
        if current_row_count + len(chunk) > chunk_size:
            strict = True
            if strict:
                needed_length = chunk_size - current_row_count
                current_row_count += needed_length
                assert current_row_count == chunk_size


                chunks_head = {name: vaex.array_types.slice(chunk, 0, needed_length) for name, chunk in chunks_current.items()}
                chunks_current_list.append(chunks_head)
                chunks_extra = {name: vaex.array_types.slice(chunk, needed_length) for name, chunk in chunks_current.items()}
                chunks_ready_list.insert(0, chunks_extra)  # put back the extra in front
            else:
                current_row_count += len(chunk)
                chunks_current_list.append(chunks_current)
        else:
            current_row_count += len(chunk)
            chunks_current_list.append(chunks_current)
    return chunks_current_list, current_row_count


def chunk_rechunk(chunk_iter, chunk_size):
    chunks_ready_list = []
    i1 = i2 = 0
    for _, _, chunks in chunk_iter:
        chunks_ready_list.append(chunks)
        total_row_count = sum([len(list(k.values())[0]) for k in chunks_ready_list])
        if total_row_count > chunk_size:
            chunks_current_list, current_row_count = vaex.dataset._slice_of_chunks(chunks_ready_list, chunk_size)
            i2 += current_row_count
            chunks = vaex.dataset._concat_chunk_list(chunks_current_list)
            yield i1, i2, chunks
            i1 = i2

    while chunks_ready_list:
        chunks_current_list, current_row_count = vaex.dataset._slice_of_chunks(chunks_ready_list, chunk_size)
        i2 += current_row_count
        chunks = vaex.dataset._concat_chunk_list(chunks_current_list)
        yield i1, i2, chunks
        i1 = i2


def _rechunk(chunk_iter, chunk_size):
    def wrapper():
        i1 = i2 = 0
        for chunks in chunk_iter:
            i2 += len(list(chunks.values())[0])
            yield i1, i2, chunks
            i1 = i2
    yield from chunk_rechunk(wrapper(), chunk_size)


def empty_chunk_iterator(start, end, chunk_size):
    length = end - start
    i1 = 0
    i2 = min(length, i1 + chunk_size)
    while i1 < length:
        yield i1, i2, {}
        i1 = i2
        i2 = min(length, i1 + chunk_size)


class Dataset(collections.abc.Mapping):
    def __init__(self):
        super().__init__()
        self._columns = frozendict()
        self._row_count = None
        self._id = str(uuid.uuid4())
        self._cached_fingerprint = None

    def _check_existence(self, column_name):
        matches = difflib.get_close_matches(column_name, list(self._columns))
        msg = "Column or variable %r does not exist." % column_name
        if matches:
            msg += ' Did you mean: ' + " or ".join(map(repr, matches))
        else:
            msg += ' Available columns or variables: ' + ", ".join(map(repr, list(self._columns)))
        raise KeyError(msg)

    def __repr__(self):
        import yaml
        data = self.__repr_data__()
        return yaml.dump(data, sort_keys=False, indent=4)

    def __repr_data__(self):
        state = self.__getstate__()
        def normalize(v):
            if isinstance(v, Dataset):
                return v.__repr_data__()
            if isinstance(v, frozendict):
                return dict(v)
            if isinstance(v, vaex.dataframe.DataFrame):
                return {'type': 'dataframe', 'repr': repr(v)}
            if isinstance(v, np.ndarray):
                return v.tolist()
            return v
        return {'type': self.snake_name, **{k: normalize(v) for k, v in state.items() if not k.startswith('_')}}

    @property
    def id(self):
        '''id that uniquely identifies a dataset at runtime'''
        return self.fingerprint

    @property
    def fingerprint(self):
        '''id that uniquely identifies a dataset cross runtime, might be more expensive and require hasing'''
        if self._cached_fingerprint is None:
            self._cached_fingerprint = self._fingerprint
        return self._cached_fingerprint

    @abstractproperty
    def _fingerprint(self):
        pass

    def encode(self, encoding):
        if not encoding.has_object_spec(self.id):
            spec = self._encode(encoding)
            encoding.set_object_spec(self.id, spec)
        return {'dataset_type': self.snake_name, 'object-id': self.id}

    @classmethod
    def decode(cls, encoding, spec):
        id = spec['object-id']
        if not encoding.has_object(id):
            spec = encoding.get_object_spec(id)
            ds = cls._decode(encoding, spec)
            encoding.set_object(id, ds)
        return encoding.get_object(id)

    @abstractmethod
    def _create_columns(self):
        pass

    @property
    def name(self):
        # TODO: in the future, we might want to use self.fingerprint or self.id
        return "no-name"

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_columns']
        del state['_cached_fingerprint']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state.copy())
        self._cached_fingerprint = None
        self._create_columns()

    def schema(self, array_type=None):
        return {name: vaex.array_types.data_type(col) for name, col in self.items()}

    def shapes(self):
        return {name: self.shape(name) for name, col in self.items()}

    def _set_row_count(self):
        if not self._columns:
            return
        values = list(self._columns.values())
        self._row_count = len(values[0])
        for name, value in list(self._columns.items())[1:]:
            if len(value) != self._row_count:
                raise ValueError(f'First columns has length {self._row_count}, while column {name} has length {len(value)}')

    @property
    def row_count(self):
        return self._row_count

    def project(self, *names):
        all = set(self)
        drop = all - set(names)
        # we want a deterministic order for fingerprints
        drop = list(drop)
        drop.sort()
        return self.dropped(*list(drop))

    def concat(self, *others, resolver='flexible'):
        datasets = []
        if isinstance(self, DatasetConcatenated):
            datasets.extend(self.datasets)
        else:
            datasets.extend([self])
        for other in others:
            if isinstance(other, DatasetConcatenated):
                datasets.extend(other.datasets)
            else:
                datasets.extend([other])
        return DatasetConcatenated(datasets, resolver=resolver)

    def take(self, indices, masked=False):
        return DatasetTake(self, indices, masked=masked)

    def renamed(self, renaming):
        return DatasetRenamed(self, renaming)

    def merged(self, rhs):
        return DatasetMerged(self, rhs)

    def dropped(self, *names):
        return DatasetDropped(self, names)

    def __getitem__(self, item):
        if isinstance(item, slice):
            assert item.step in [1, None]
            return self.slice(item.start or 0, item.stop or self.row_count)
        return self._columns[item]

    def __len__(self):
        return len(self._columns)

    def __iter__(self):
        return iter(self._columns)

    def get_data(self, i1, i2, names):
        raise NotImplementedError

    def __eq__(self, rhs):
        if not isinstance(rhs, Dataset):
            return NotImplemented
        # simple case, if fingerprints are equal, the data is equal
        
        if self.fingerprint == rhs.fingerprint:
            return True
        # but no the other way around
        keys = set(self)
        keys_hashed = set(self._ids)
        missing = keys ^ keys_hashed
        if missing:
            return self.fingerprint == rhs.fingerprint
        keys = set(rhs)
        keys_hashed = set(rhs._ids)
        missing = keys ^ keys_hashed
        if missing:
            return self.fingerprint == rhs.fingerprint
        return self._ids == rhs._ids

    def __hash__(self):
        keys = set(self)
        keys_hashed = set(self._ids)
        missing = keys ^ keys_hashed
        if missing:
            # if we don't have hashes for all columns, we just use the fingerprint
            return hash(self.fingerprint)
        return hash(tuple(self._ids.items()))

    def _default_lazy_chunk_iterator(self, array_map, columns, chunk_size, reverse=False):
        chunk_size = chunk_size or 1024**2
        chunk_count = (self.row_count + chunk_size - 1) // chunk_size
        chunks = range(chunk_count)
        if reverse:
            chunks = reversed(chunks)
        for i in chunks:
            i1 = i * chunk_size
            i2 = min((i + 1) * chunk_size, self.row_count)
            def reader(i1=i1, i2=i2):
                chunks = {k: array_map[k][i1:i2] for k in columns}
                length = i2 - i1
                for name, chunk in chunks.items():
                    assert len(chunk) == length, f'Oops, got a chunk ({name}) of length {len(chunk)} while it is expected to be of length {length} (at {i1}-{i2}'
                return chunks
            yield i1, i2, reader

    def _default_chunk_iterator(self, array_map, columns, chunk_size, reverse=False):
        for i1, i2, reader in self._default_lazy_chunk_iterator(array_map, columns, chunk_size, reverse):
            yield i1, i2, reader()

    @abstractmethod
    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        pass

    @abstractmethod
    def is_masked(self, column):
        pass

    @abstractmethod
    def shape(self, column):
        pass

    @abstractmethod
    def close(self):
        '''Close file handles or other resources, the DataFrame will not be in a usable state afterwards.'''
        pass

    @abstractmethod
    def slice(self, start, end):
        pass

    @abstractmethod
    def hashed(self):
        pass

    @abstractmethod
    def leafs(self) -> List["Dataset"]:
        pass


class DatasetDecorator(Dataset):
    def __init__(self, original):
        super().__init__()
        self.original = original

    def leafs(self) -> List[Dataset]:
        return self.original.leafs()

    def close(self):
        self.original.close()

    def is_masked(self, column):
        return self.original.is_masked(column)

    def shape(self, column):
        return self.original.shape(column)


class ColumnProxy(vaex.column.Column):
    '''To give the Dataset._columns object useful containers for debugging'''
    ds: Dataset

    def __init__(self, ds, name, type):
        self.ds = ds
        self.name = name
        self.dtype = type

    def _fingerprint(self):
        fp = vaex.cache.fingerprint(self.ds.fingerprint, self.name)
        return f'column-proxy-{fp}'

    def __len__(self):
        return self.ds.row_count

    def to_numpy(self):
        values = self[:]
        return np.array(values)

    def __getitem__(self, item):
        if isinstance(item, slice):
            array_chunks = []
            ds = self.ds.__getitem__(item)
            for chunk_start, chunk_end, chunks in ds.chunk_iterator([self.name]):
                ar = chunks[self.name]
                if isinstance(ar, pa.ChunkedArray):
                    array_chunks.extend(ar.chunks)
                else:
                    array_chunks.append(ar)
            if len(array_chunks) == 1:
                return array_chunks[0]
            if len(array_chunks) == 0:
                return vaex.dtype(self.dtype).create_array([])
            return vaex.array_types.concat(array_chunks)
        else:
            raise NotImplementedError


@register
class DatasetRenamed(DatasetDecorator):
    snake_name = 'rename'
    def __init__(self, original, renaming):
        super().__init__(original)
        self.renaming = renaming
        self.reverse = {v: k for k, v in renaming.items()}
        self._create_columns()
        self._ids = frozendict({renaming.get(name, name): ar for name, ar in original._ids.items()})
        self._set_row_count()

    def renamed(self, renaming):
        # # {'a': 'x', 'b': 'y'} and {'x': 'a', 'b': 'z', 'c', 'q'} -> {'b': 'z', 'c': 'q'}
        resulting = {}
        renaming = renaming.copy()  # we'll modify in place
        for old, new in self.renaming.items():
            if new in renaming:
                if old == renaming[new]:
                    pass # e.g.  x->a->x
                else:
                    resulting[old] = renaming[new]
                del renaming[new]  # we already covered this
            else:
                # e.g. x->a->a
                resulting[old] = new
        # e.g. x->x->a
        resulting.update(renaming)
        return DatasetRenamed(self.original, resulting)

    @property
    def _fingerprint(self):
        id = vaex.cache.fingerprint(self.original.fingerprint, self.renaming)
        return f'dataset-{self.snake_name}-{self.original.fingerprint}'

    def _create_columns(self):
        self._columns = frozendict({self.renaming.get(name, name): ar for name, ar in self.original.items()})

    def _encode(self, encoding):
        dataset_spec = encoding.encode('dataset', self.original)
        return {'renaming': dict(self.renaming), 'dataset': dataset_spec}

    @classmethod
    def _decode(cls, encoding, spec):
        dataset = encoding.decode('dataset', spec['dataset'])
        return cls(dataset, spec['renaming'])

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        for name in columns:
            if name in self.renaming:
                rename = self.renaming[name]
                raise KeyError(f'Oops, you tried to get column {name}, but you renamed it to {rename}')
        columns = [self.reverse.get(name, name) for name in columns]
        for i1, i2, chunks in self.original.chunk_iterator(columns, chunk_size, reverse=reverse):
            yield i1, i2, {self.renaming.get(name, name): ar for name, ar in chunks.items()}

    def is_masked(self, column):
        return self.original.is_masked(self.reverse.get(column, column))

    def shape(self, column):
        return self.original.shape(self.reverse.get(column, column))

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return type(self)(self.original.slice(start, end), self.renaming)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self.renaming)


@register
class DatasetConcatenated(Dataset):
    snake_name = "concat"
    def __init__(self, datasets, resolver):
        super().__init__()
        self.datasets = datasets
        self.resolver = resolver
        if self.resolver == 'strict':
            for dataset in datasets[1:]:
                if set(dataset) != set(datasets[0]):
                    l = set(dataset)
                    r = set(datasets[0])
                    diff = l ^ r
                    raise NameError(f'Concatenating datasets with different names: {l} and {r} (difference: {diff})')
            self._schema = datasets[0].schema()
            self._shapes = datasets[0].shapes()
            for dataset in datasets[1:]:
                if dataset.shapes() != self._shapes:
                    raise ValueError(f'Cannot concatenate with different shapes: {self._shapes} != {dataset.shapes()}')
            for dataset in datasets[1:]:
                schema = dataset.schema()
                if dataset.schema() != self._schema:
                    raise ValueError(f'Cannot concatenate with different schemas: {self._shapes} != {dataset.shapes()}')
        elif self.resolver == 'flexible':
            schemas = [ds.schema() for ds in datasets]
            shapes = [ds.shapes() for ds in datasets]
            # try to keep the order of the original dataset
            schema_list_map = {}
            for schema in schemas:
                for name, type in schema.items():
                    if name not in schema_list_map:
                        schema_list_map[name] = []
            for name, type_list in schema_list_map.items():
                for schema in schemas:
                    # None means it is means the column is missing
                    type_list.append(schema.get(name))
            from .schema import resolver_flexible

            # shapes
            shape_list_map = {}
            for shape in shapes:
                for name, type in shape.items():
                    if name not in shape_list_map:
                        shape_list_map[name] = []
            for name, shape_list in shape_list_map.items():
                for shapes_ in shapes:
                    # None means it is means the column is missing
                    shape_list.append(shapes_.get(name))
            self._schema = {}
            self._shapes = {}

            for name in shape_list_map:
                self._schema[name], self._shapes[name] = resolver_flexible.resolve(schema_list_map[name], shape_list_map[name])
        else:
            raise ValueError(f'Invalid resolver {resolver}, choose between "strict" or "flexible"')

        self._create_columns()
        self._set_row_count()

    @property
    def _fingerprint(self):
        ids = [ds.fingerprint for ds in self.datasets]
        id = vaex.cache.fingerprint(*ids)
        return f'dataset-{self.snake_name}-{id}'

    def _create_columns(self):
        columns = {}
        hashes = {}
        for name in self._schema:
            columns[name] = ColumnProxy(self, name, self._schema[name])
            if all(name in ds._ids for ds in self.datasets):
                hashes[name] = hash_combine(*[ds._ids[name] for ds in self.datasets])
        self._columns = frozendict(columns)
        self._ids = frozendict(hashes)

    def _encode(self, encoding, skip=set()):
        datasets = encoding.encode_list('dataset', self.datasets)
        spec = {'dataset_type': self.snake_name, 'datasets': datasets, 'resolver': self.resolver}
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        datasets = encoding.decode_list('dataset', spec['datasets'])
        ds = cls(datasets, spec['resolver'])
        return ds

    def is_masked(self, column):
        for dataset in self.datasets:
            if column not in dataset:
                # if the column is not in the dataset, we assume it is not masked
                # since we ues arrow in that case
                return False
        return any(k.is_masked(column) for k in self.datasets)

    def shape(self, column):
        return self._shapes[column]

    def _set_row_count(self):
        self._row_count = sum(ds.row_count for ds in self.datasets)

    def schema(self, array_type=None):
        return self._schema.copy()

    def _chunk_iterator_non_strict(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        end = self.row_count if end is None else end
        offset = 0
        for dataset in self.datasets:
            present = [k for k in columns if k in dataset]
            # skip over whole datasets
            if start >= offset + dataset.row_count:
                offset += dataset.row_count
                continue
            # we are past the end
            if end <= offset:
                break
            for i1, i2, chunks in dataset.chunk_iterator(present, chunk_size=chunk_size, reverse=reverse):
                # chunks = {name: vaex.array_types.to_arrow(ar) for name, ar in chunks.items()}
                length = i2 - i1
                chunk_start = offset
                chunk_end = offset + length
                if start >= chunk_end:  # we didn't find the beginning yet
                    offset += length
                    continue
                if end <= chunk_start:  # we are past the end
                    # assert False
                    break

                if start > chunk_start:
                    # this means we have to cut off a piece of the beginning
                    if end < chunk_end:
                        # AND the end
                        length = end - chunk_start  # without the start cut off
                        length -= start - chunk_start  # correcting for the start cut off
                        assert length > 0
                        chunks = {name: vaex.array_types.slice(ar, start - chunk_start, length) for name, ar in chunks.items()}
                        for name, ar in chunks.items():
                            assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'
                    else:
                        length -= start - chunk_start  # correcting for the start cut off
                        assert length > 0
                        chunks = {name: vaex.array_types.slice(ar, start - chunk_start) for name, ar in chunks.items()}
                        for name, ar in chunks.items():
                            assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'
                else:
                    if end < chunk_end:
                        # we only need to cut off a piece of the end
                        length = end - chunk_start
                        assert length > 0
                        chunks = {name: vaex.array_types.slice(ar, 0, length) for name, ar in chunks.items()}
                        for name, ar in chunks.items():
                            assert len(ar) == length, f'Oops, array was expected to be of length {length} but was {len(ar)}'

                from .schema import resolver_flexible
                allchunks = {name: resolver_flexible.align(length, chunks.get(name), self._schema[name], self._shapes[name]) for name in columns}
                yield {k: allchunks[k] for k in columns}
                offset += (i2 - i1)

    def chunk_iterator(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        chunk_size = chunk_size or 1024*1024
        i1 = 0
        i1 = i2 = 0
        if not columns:
            end = self.row_count if end is None else end
            yield from empty_chunk_iterator(start, end, chunk_size)
        else:
            chunk_iterator = self._chunk_iterator_non_strict(columns, chunk_size, reverse=reverse, start=start, end=self.row_count if end is None else end)
            yield from _rechunk(chunk_iterator, chunk_size)

    def close(self):
        for ds in self.datasets:
            ds.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        # TODO: we can be smarter here, and trim off some datasets
        return DatasetSliced(self, start=start, end=end)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)([dataset.hashed() for dataset in self.datasets], resolver=self.resolver)

    def leafs(self) -> List[Dataset]:
        return [self]

    # def leafs(self) -> List[Dataset]:
    #     leafs = list()
    #     for ds in self.datasets:
    #         leafs.extend(ds.leafs())
    #     return leafs

@register
class DatasetTake(DatasetDecorator):
    snake_name = "take"
    def __init__(self, original, indices, masked):
        super().__init__(original)
        self.indices = indices
        self.masked = masked
        self._lazy_hash_index = None
        self._create_columns()
        self._set_row_count()

    @property
    def _fingerprint(self):
        id = vaex.cache.fingerprint(self.original.fingerprint, self._hash_index, self.masked)
        return f'dataset-{self.snake_name}-{id}'

    @property
    def _hash_index(self):
        if self._lazy_hash_index is None:
            self._lazy_hash_index = hash_array(self.indices)
        return self._lazy_hash_index

    def _create_columns(self):
        # if the columns in ds already have a ColumnIndex
        # we could do, direct_indices = df.column['bla'].indices[indices]
        # which should be shared among multiple ColumnIndex'es, so we store
        # them in this dict
        direct_indices_map = {}
        columns = {}
        hashes = {}
        for name, column in self.original.items():
            columns[name] = ColumnIndexed.index(column, self.indices, direct_indices_map, masked=self.masked)
            if name in self.original._ids:
                hashes[name] = hash_combine(self._hash_index, self.original._ids[name])
        self._columns = frozendict(columns)
        self._ids = frozendict(hashes)

    def _encode(self, encoding, skip=set()):
        dataset_spec = encoding.encode('dataset', self.original)
        spec = {'dataset_type': self.snake_name, 'dataset': dataset_spec}
        spec['indices'] = encoding.encode('array', self.indices)
        spec['masked'] = self.masked
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        dataset = encoding.decode('dataset', spec['dataset'])
        indices = encoding.decode('array', spec['indices'])
        ds = cls(dataset, indices, spec['masked'])
        return ds

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        # TODO: we may be able to do this slightly more efficient by first
        # materializing the columns
        yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return DatasetSlicedArrays(self, start=start, end=end)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self.indices, self.masked)


@register
class DatasetFiltered(DatasetDecorator):
    snake_name = 'filter'
    def __init__(self, original, filter, expected_length=None, state=None, selection=None):
        super().__init__(original)
        self._filter = filter
        self._lazy_hash_filter = None
        self._create_columns()
        self._row_count = np.sum(self._filter).item()
        self.state = state
        self.selection = selection
        if expected_length is not None:
            if expected_length != self._row_count:
                raise ValueError(f'Expected filter to have {expected_length} true values, but counted {self._row_count}')

    @property
    def _fingerprint(self):
        id = vaex.cache.fingerprint(self.original.id, self._hash_index, self.state, self.selection)
        return f'dataset-{self.snake_name}-{id}'

    @property
    def _hash_index(self):
        if self._lazy_hash_filter is None:
            self._lazy_hash_filter = hash_array(self._filter)
        return self._lazy_hash_filter

    def _create_columns(self):
        columns = {name: vaex.dataset.ColumnProxy(self, name, data_type(col)) for name, col in self.original._columns.items()}
        hashes = {}
        for name, column in self.original.items():
            if name in self.original._ids:
                hashes[name] = hash_combine(self._hash_index, self.original._ids[name])
        self._columns = frozendict(columns)
        self._ids = frozendict(hashes)

    def _encode(self, encoding, skip=set()):
        dataset_spec = encoding.encode('dataset', self.original)
        spec = {'dataset': dataset_spec}
        if self.state is not None and self.selection is not None:
            spec['state'] = encoding.encode('dataframe-state', self.state)
            spec['selection'] = encoding.encode('selection', self.selection)
        spec['filter_array'] = encoding.encode('array', self._filter)
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        dataset = encoding.decode('dataset', spec['dataset'])
        if 'filter_array' in spec:
            filter = encoding.decode('array', spec['filter_array'])
            ds = cls(dataset, filter)
        else:
            state = encoding.decode('dataframe-state', spec['state'])
            selection = encoding.decode('selection', spec['selection'])
            df = vaex.from_dataset(dataset)
            df.state_set(state)
            df.set_selection(vaex.dataframe.FILTER_SELECTION_NAME, selection)
            df._push_down_filter()
            filter = df.dataset.filter
            ds = cls(dataset, filter, state=state, selection=selection)
        return ds

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        chunk_size = chunk_size or 1024**2
        if not columns:
            end = self.row_count
            length = end
            i1 = i2 = 0
            i2 = min(length, i1 + chunk_size)
            while i1 < length:
                yield i1, i2, {}
                i1 = i2
                i2 = min(length, i1 + chunk_size)
            return
        def filtered_chunks():
            for i1, i2, chunks in self.original.chunk_iterator(columns, chunk_size=chunk_size, reverse=reverse):
                chunks_filtered = {name: vaex.array_types.filter(ar, self._filter[i1:i2]) for name, ar in chunks.items()}
                yield chunks_filtered
        yield from _rechunk(filtered_chunks(), chunk_size)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self._filter)

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        expected_length = end - start
        mask = vaex.superutils.Mask(memoryview(self._filter))
        start, end = mask.indices(start, end-1)
        end += 1
        filter = self._filter[start:end]
        assert filter.sum() == expected_length
        return type(self)(self.original.slice(start, end), filter)

@register
class DatasetSliced(DatasetDecorator):
    snake_name = "slice"
    def __init__(self, original, start, end):
        super().__init__(original)
        self.start = start
        self.end = end
        self._row_count = end - start
        self._create_columns()
        # self._ids = {}
        self._ids = frozendict({name: hash_slice(hash, start, end) for name, hash in original._ids.items()})

    @property
    def _fingerprint(self):
        id = vaex.cache.fingerprint(self.original.fingerprint, self.start, self.end)
        return f'dataset-{self.snake_name}-{id}'

    def leafs(self) -> List[Dataset]:
        # we don't want to propagate slicing
        return [self]

    def _encode(self, encoding, skip=set()):
        dataset_spec = encoding.encode('dataset', self.original)
        return {'dataset': dataset_spec, 'start': self.start, 'end': self.end}

    @classmethod
    def _decode(cls, encoding, spec):
        dataset = encoding.decode('dataset', spec['dataset'])
        return cls(dataset, spec['start'], spec['end'])

    def _create_columns(self):
        self._columns = {name: vaex.dataset.ColumnProxy(self, name, data_type(col)) for name, col in self.original._columns.items()}

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        yield from self.original.chunk_iterator(columns, chunk_size=chunk_size, reverse=reverse, start=self.start, end=self.end)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self.start, self.end)

    def slice(self, start, end):
        length = end - start
        start += self.start
        end = start + length
        if end > self.original.row_count:
            raise IndexError(f'Slice end ({end}) if larger than number of rows: {self.original.row_count}')
        return type(self)(self.original, start, end)


@register
class DatasetSlicedArrays(DatasetDecorator):
    snake_name = 'slice_arrays'
    def __init__(self, original, start, end):
        super().__init__(original)
        # maybe we want to avoid slicing twice, and collapse it to 1?
        self.start = start
        self.end = end
        # TODO: this is the old dataframe.trim method, we somehow need to test/capture that
        # if isinstance(column, array_types.supported_array_types):  # real array
        #     df.columns[name] = column[self._index_start:self._index_end]
        # else:
        #     df.columns[name] = column.trim(self._index_start, self._index_end)
        self._create_columns()
        self._ids = frozendict({name: hash_slice(hash, start, end) for name, hash in original._ids.items()})
        self._set_row_count()

    @property
    def _fingerprint(self):
        id = vaex.cache.fingerprint(self.original.fingerprint, self.start, self.end)
        return f'dataset-{self.snake_name}-{id}'

    def leafs(self) -> List[Dataset]:
        # we don't want to propagate slicing
        return [self]

    def _create_columns(self):
        columns = {}
        for name, column in self.original.items():
            if isinstance(column, array_types.supported_array_types):  # real array
                column = column[self.start:self.end]
            else:
                column = column.trim(self.start, self.end)
            columns[name] = column
        self._columns = frozendict(columns)

    def _encode(self, encoding, skip=set()):
        dataset_spec = encoding.encode('dataset', self.original)
        return {'dataset': dataset_spec, 'start': self.start, 'end': self.end}

    @classmethod
    def _decode(cls, encoding, spec):
        dataset = encoding.decode('dataset', spec['dataset'])
        return cls(dataset, spec['start'], spec['end'])

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self.start, self.end)

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        length = end - start
        start += self.start
        end = start + length
        if end > self.original.row_count:
            raise IndexError(f'Slice end ({end}) if larger than number of rows: {self.original.row_count}')
        return type(self)(self.original, start, end)


@register
class DatasetDropped(DatasetDecorator):
    snake_name = "drop"
    def __init__(self, original, names):
        super().__init__(original)
        self._dropped_names = tuple(names)
        self._create_columns()
        self._ids = frozendict({name: ar for name, ar in original._ids.items() if name not in names})
        self._set_row_count()

    def dropped(self, *names):
        return DatasetDropped(self.original, self._dropped_names + names)

    @property
    def _fingerprint(self):
        id = vaex.cache.fingerprint(self.original.fingerprint, self._dropped_names)
        return f'dataset-{self.snake_name}-{id}'

    def _create_columns(self):
        self._columns = frozendict({name: ar for name, ar in self.original.items() if name not in self._dropped_names})

    def _encode(self, encoding):
        dataset_spec = encoding.encode('dataset', self.original)
        return {'dataset': dataset_spec, 'names': list(self._dropped_names)}

    @classmethod
    def _decode(cls, encoding, spec):
        dataset = encoding.decode('dataset', spec['dataset'])
        ds = cls(dataset, spec['names'])
        return ds

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        for column in columns:
            if column in self._dropped_names:
                raise KeyError(f'Oops, you tried to get column {column} while it is actually dropped')
        yield from self.original.chunk_iterator(columns, chunk_size=chunk_size, reverse=reverse)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self._dropped_names)

    def close(self):
        self.original.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return type(self)(self.original.slice(start, end), self._dropped_names)


@register
class DatasetMerged(Dataset):
    snake_name = "merge"
    def __init__(self, left, right):
        super().__init__()
        self.left = left
        self.right = right
        if self.left.row_count != self.right.row_count:
            raise ValueError(f'Merging datasets with unequal row counts ({self.left.row_count} != {self.right.row_count})')
        self._row_count = self.left.row_count
        overlap = set(left) & set(right)
        if overlap:
            raise NameError(f'Duplicate names: {overlap}')
        self._create_columns()
        self._ids = frozendict({**left._ids, **right._ids})
        self._set_row_count()

    @property
    def _fingerprint(self):
        id = vaex.cache.fingerprint(self.left.fingerprint, self.right.fingerprint)
        return f'dataset-{self.snake_name}-{id}'

    def leafs(self) -> List[Dataset]:
        return self.left.leafs() + self.right.leafs()

    def _create_columns(self):
        # TODO: for DatasetArray, we might want to just do this?
        # self._columns = frozendict({**left._columns, **right._columns})
        self._columns = {**{name: ColumnProxy(self.left, name, data_type(col)) for name, col in self.left._columns.items()},
                         **{name: ColumnProxy(self.right, name, data_type(col)) for name, col in self.right._columns.items()}}

    def _encode(self, encoding, skip=set()):
        dataset_spec_left = encoding.encode('dataset', self.left)
        dataset_spec_right = encoding.encode('dataset', self.right)
        spec = {'left': dataset_spec_left, 'right': dataset_spec_right}
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        left = encoding.decode('dataset', spec['left'])
        right = encoding.decode('dataset', spec['right'])
        ds = cls(left, right)
        return ds

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        columns_left = [k for k in columns if k in self.left]
        columns_right = [k for k in columns if k in self.right]
        if not columns_left:
            yield from self.right.chunk_iterator(columns, chunk_size, reverse=reverse)
        elif not columns_right:
            yield from self.left.chunk_iterator(columns, chunk_size, reverse=reverse)
        else:
            for (i1, i2, ichunks), (j1, j2, jchunks) in zip(
                self.left.chunk_iterator(columns_left, chunk_size, reverse=reverse),
                self.right.chunk_iterator(columns_right, chunk_size, reverse=reverse)):
                # TODO: if one of the datasets does not respect the chunk_size (e.g. parquet)
                # this might fail
                assert i1 == j1
                assert i2 == j2
                yield i1, i2, {**ichunks, **jchunks}

    def is_masked(self, column):
        if column in self.left:
            return self.left.is_masked(column)
        else:
            return self.right.is_masked(column)

    def shape(self, column):
        if column in self.left:
            return self.left.shape(column)
        else:
            return self.right.shape(column)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.left.hashed(), self.right.hashed())

    def close(self):
        self.left.close()
        self.right.close()

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return type(self)(self.left.slice(start, end), self.right.slice(start, end))


@register
class DatasetArrays(Dataset):
    snake_name = "arrays"
    def __init__(self, mapping=None, hashed=True, **kwargs):
        super().__init__()
        if mapping is None:
            mapping = {}
        columns = {**mapping, **kwargs}
        columns = {key: to_supported_array(ar) for key, ar in columns.items()}
        # TODO: we finally want to get rid of datasets with no columns
        self._columns = frozendict(columns)
        if hashed:
            self._ids = frozendict({key: hash_array(array) for key, array in self._columns.items()})
        else:
            self._ids = frozendict()
        self._set_row_count()

    @property
    def id(self):
        try:
            # requires hashing and is expensive
            return self.fingerprint
        except ValueError:
            return f'dataset-{self.snake_name}-uuid4-{self._id}'

    @property
    def _fingerprint(self):
        keys = set(self)
        keys_hashed = set(self._ids)
        missing = keys ^ keys_hashed
        if missing:
            # if we don't have hashes for all columns, we do it like id
            return f'dataset-{self.snake_name}-uuid4-{self._id}'
        # self.__hash__()  # invoke just to check we don't have missing hashes
        # but Python's hash functions are not deterministic (cross processs)
        fp = vaex.cache.fingerprint(tuple(self._ids.items()))
        return f'dataset-{self.snake_name}-hashed-{fp}'

    def leafs(self) -> List[Dataset]:
        return [self]

    def _encode(self, encoding):
        arrays = encoding.encode_dict('array', self._columns)
        spec = {'dataset_type': self.snake_name, 'arrays': arrays}
        if self._ids:
            fingerprints = dict(self._ids)
            spec['fingerprints'] = fingerprints
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        arrays = encoding.decode_dict('array', spec['arrays'])
        ds = cls(arrays)
        if 'fingerprints' in spec:
            ds._ids = frozendict(spec['fingerprints'])
        return ds

    def __getstate__(self):
        state = self.__dict__.copy()
        # here, we actually DO want to keep the columns
        # del state['_columns']
        return state

    def __setstate__(self, state):
        super().__setstate__(state)

    def _create_columns(self):
        pass

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def is_masked(self, column):
        ar = self._columns[column]
        if not isinstance(ar, np.ndarray):
            ar = ar[0:1]  # take a small piece
        if isinstance(ar, np.ndarray):
            return np.ma.isMaskedArray(ar)
        else:
            return False  # an arrow array always has null value options

    def shape(self, column):
        ar = self._columns[column]
        if not isinstance(ar, np.ndarray):
            ar = ar[0:1]  # take a small piece
        if isinstance(ar, vaex.array_types.supported_arrow_array_types):
            return tuple()
        else:
            return ar.shape[1:]

    def merged(self, rhs):
        # TODO: if we don't allow emtpy datasets, we can remove this method
        if len(self) == 0:
            return rhs
        if len(rhs) == 0:
            return self
        # TODO: this is where we want to check if both are array like
        # and have faster version of merged
        return DatasetMerged(self, rhs)

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return DatasetSlicedArrays(self, start=start, end=end)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        new = type(self)(self._columns)
        new._ids = frozendict({key: hash_array(array) for key, array in new._columns.items()})
        return new

    def close(self):
        pass  # nothing to do, maybe drop a refcount?

    # TODO: we might want to really get rid of these, since we want to avoid copying them over the network?
    # def dropped(self, names):

class DatasetFile(Dataset):
    """Datasets that map to a file can keep their ids/hashes in the file itself,
    or keep them in a meta file.
    """
    def __init__(self, path, write=False, fs_options={}, fs=None):
        super().__init__()
        self.path = path
        self.fs_options = fs_options
        self.fs = fs
        self.write = write
        self._columns = {}
        self._ids = {}
        self._frozen = False
        self._hash_calculations = 0  # track it for testing purposes
        self._hash_info = {}
        self._hash_cache_needs_write = False
        self._read_hashes()

    @property
    def name(self):
        base, ext, fs_options = vaex.file.split_ext(self.path)
        base = os.path.basename(base)
        return base

    @property
    def _fingerprint(self):
        if set(self._ids) == set(self):
            fingerprint = vaex.cache.fingerprint(dict(self._ids))
            return f'dataset-{self.snake_name}-hashed-{fingerprint}'
        else:
            # TODO: if the dataset is hashed, return a fingerprint based on that
            fingerprint = vaex.file.fingerprint(self.path, fs_options=self.fs_options, fs=self.fs)
            return f'dataset-{self.snake_name}-{fingerprint}'

    def leafs(self) -> List[Dataset]:
        return [self]

    def _create_columns(self):
        pass

    @classmethod
    def quick_test(cls, path, fs_options={}, fs=None, *args, **kwargs):
        return False

    @classmethod
    def open(cls, path, *args, **kwargs):
        return cls(path, *args, **kwargs)

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        yield from self._default_chunk_iterator(self._columns, columns, chunk_size, reverse=reverse)

    def is_masked(self, column):
        ar = self._columns[column]
        if not isinstance(ar, np.ndarray):
            ar = ar[0:1]  # take a small piece
        if isinstance(ar, np.ndarray):
            return np.ma.isMaskedArray(ar)
        else:
            return False  # an arrow array always has null value options

    def shape(self, column):
        ar = self._columns[column]
        if not isinstance(ar, np.ndarray):
            ar = ar[0:1]  # take a small piece
        if isinstance(ar, vaex.array_types.supported_arrow_array_types):
            return tuple()
        else:
            return ar.shape[1:]

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return DatasetSlicedArrays(self, start=start, end=end)

    def _read_hashes(self):
        path_hashes = Path(str(self.path )+ '.d') / 'hashes.yaml'
        try:
            exists = path_hashes.exists()
        except OSError:  # happens for windows py<38
            exists = False
        if exists:
            with path_hashes.open() as f:
                hashes = vaex.utils.yaml_load(f)
                if hashes is None:
                    raise ValueError(f'{path_hashes} was probably truncated due to another process writing.')
                self._hash_info = hashes.get('columns', {})

    def _freeze(self):
        self._ids = frozendict(self._ids)
        self._columns = frozendict(self._columns)
        self._set_row_count()
        self._frozen = True
        if self._hash_cache_needs_write:
            self._write_hash_info()

    def _encode(self, encoding):
        spec = {'write': self.write,
                'path': str(self.path),
                'fs_options': self.fs_options,
                'fs': self.fs}
        return spec

    @classmethod
    def _decode(cls, encoding, spec):
        return cls(**spec)

    def __getstate__(self):
        # we don't have the columns in the state, since we should be able
        # to get them from disk again
        state = {
            'write': self.write,
            'path': self.path,
            'fs_options': self.fs_options,
            'fs': self.fs,
            '_ids': dict(self._ids)  # serialize the hases as non-frozen dict
        }
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        # 'ctor' like initialization
        self._frozen = False
        self._hash_calculations = 0
        self._columns = {}
        self._hash_info = {}
        self._hash_cache_needs_write = False
        self._read_hashes()

    def add_column(self, name, data):
        self._columns[name] = data
        if self.write:
            return  # the columns don't include the final data
            # the hashes will be done in .freeze()
        hash_info = self._hash_info.get(name)
        if hash_info:
            hash_info_previous = hash_info.copy()
            hash, hash_info = hash_array(data, hash_info, return_info=True)
            if hash_info_previous != hash_info:
                self._hash_cache_needs_write = True
            self._ids[name] = hash
            self._hash_info[name] = hash_info  # always update the information

    @property
    def _local_hash_path(self):
        # TODO: support s3 and gcs
        # TODO: fallback directory when a user cannot write
        if Path(self.path).exists():
            directory = Path(self.path + '.d')
            directory.mkdir(exist_ok=True)
        else:
            o = urlparse(self.path)
            directory = Path(vaex.utils.get_private_dir('dataset', o.scheme, o.netloc, o.path[1:]))
        return directory / 'hashes.yaml'

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        cls = type(self)
        # use pickle protocol to clone
        new = cls.__new__(cls)
        new.__setstate__(self.__getstate__())
        hashes = {}
        disk_cached_hashes = {}
        for name, column in new.items():
            hash_info = self._hash_info.get(name)
            if hash_info is None:
                logging.warning(f'Calculating hash for column {name} of length {len(column)} (1 time operation, will be cached on disk)')
                hash_info = hash_array_data(column)
            hash, hash_info = hash_array(column, hash_info, return_info=True)
            new._hash_calculations += 1
            hashes[name] = hash
            disk_cached_hashes[name] = hash_info
        new._ids = frozendict(hashes)
        new._hash_info = frozendict(disk_cached_hashes)
        path_hashes = new._local_hash_path
        # TODO: without this check, if multiple processes are writing (e.g. tests/execution_test.py::test_task_sum with ray)
        # this leads to a race condition, where we write the file, and while truncated, _read_hases() fails (because the file exists)
        # if new._hash_info != new._ids:
        new._write_hash_info()
        return new

    def _write_hash_info(self):
        if self._hash_info:  # TODO: file lock
            path_hashes = self._local_hash_path
            with path_hashes.open('w') as f:
                vaex.utils.yaml_dump(f, {'columns': dict(self._hash_info)})


class DatasetCached(DatasetDecorator):
    snake_name = "cached"
    shared_cache = {}

    def __init__(self, original, names, cache=None, to_numpy=False):
        super(DatasetCached, self).__init__(original)
        self.original = original
        self.names = names
        self._shared = cache is None or cache is self.shared_cache
        self.cache = cache if cache is not None else self.shared_cache
        self.to_numpy = to_numpy
        self._create_columns()
        self._row_count = self.original.row_count

    @property
    def _fingerprint(self):
        return self.original.fingerprint

    def _create_columns(self):
        columns = {}
        schema = self.original.schema()
        for name, column in self.original.items():
            columns[name] = ColumnProxy(self, name, schema[name])
        self._columns = frozendict(columns)
        self._ids = frozendict(self.original._ids)

    def _encode(self, encoding, skip=set()):
        raise NotImplementedError("cannot serialize cache")

    @classmethod
    def _decode(cls, encoding, spec):
        raise NotImplementedError("cannot serialize cache")

    def chunk_iterator(self, columns, chunk_size=None, reverse=False):
        chunk_size = chunk_size or chunk_size_default
        columns_all = set(columns)
        columns_cachable = columns_all & set(self.names)
        # avoids asking the cache twice, by using .get() and then testing for None
        columns_cached = {name: self.cache.get(self._cache_key(name)) for name in columns_cachable}
        columns_cached = {name: array for name, array in columns_cached.items() if array is not None}
        columns_to_cache = columns_cachable - set(columns_cached)
        column_required = columns_all - set(columns_cached)
        cache_chunks = {name: [] for name in columns_to_cache}

        def cached_iterator():
            chunks_list = [chunks for name, chunks in columns_cached.items()]
            # chunks_list is of form [[ar1x, ar2x, a3x], [ar1y, ar2y, a3y]]
            # and now we want to yield
            #  * i1, i2 {'x': ar1x, 'y': ar1y}
            #  * i1, i2 {'x': ar2x, 'y': ar2y}
            #  * i1, i2 {'x': ar3x, 'y': ar3y}
            names = [name for name, chunks in columns_cached.items()]
            i1 = 0
            i2 = 0
            for chunks in zip(*chunks_list):
                i2 += len(chunks[0])
                for chunk in chunks:
                    assert len(chunk) == len(chunks[0])
                yield i1, i2, dict(zip(names, chunks))
                i1 = i2

        if columns_cached:
            cached_iter = chunk_rechunk(cached_iterator(), chunk_size)
        else:
            cached_iter = empty_chunk_iterator(0, self.row_count, chunk_size)
        if column_required:
            original_iter = self.original.chunk_iterator(column_required, chunk_size, reverse=reverse)
        else:
            original_iter = empty_chunk_iterator(0, self.row_count, chunk_size)
        original_iter = list(original_iter)
        cached_iter = list(cached_iter)

        for (o1, o2, ochunks), (c1, c2, cchunks) in zip(original_iter, cached_iter):
            assert o1 == c1
            assert o2 == c2
            yield o1, o2, {**ochunks, **cchunks}
            for name in columns_to_cache:
                if self.to_numpy:
                    ochunks = {k: vaex.array_types.to_numpy(v) for k, v in ochunks.items()}
                cache_chunks[name].append(ochunks[name])
        # we write it too the cache in 1 go
        for name in columns_to_cache:
            self.cache[self._cache_key(name)] = cache_chunks[name]

    def slice(self, start, end):
        if start == 0 and end == self.row_count:
            return self
        return type(self)(self.original.slice(start, end), self.names, cache=self.cache)

    def hashed(self):
        if set(self._ids) == set(self):
            return self
        return type(self)(self.original.hashed(), self.names, cache=self.cache)

    def _cache_key(self, name):
        return f"{self.fingerprint}-{name}"



'''(Currently experimental, use at own risk)
Vaex can cache task results, such as aggregations, or the internal hashmaps used for groupby to make recurring calculations much faster, at the cost of calculating cache keys and storing/retrieving the cached values.

Internally, Vaex calculates fingerprints (such as hashes of data, or file paths and mtimes) to create cache keys that are similar across processes, such that a restart of a process will most likely result in similar hash keys.

Caches can turned on globally, or used as a context manager:

>>> import vaex
>>> df = vaex.example()
>>> vaex.cache.memory_infinite()  # cache on globally
<cache restore context manager>
>>> vaex.cache.is_on()
True
>>> vaex.cache.off() # cache off globally
<cache restore context manager>
>>> vaex.cache.is_on()
False
>>> with vaex.cache.memory_infinite():
...     df.x.sum()  # calculated without cache
array(-20884.64307324)
>>> vaex.cache.is_on()
False


The functions :func:`vaex.cache.set` and  :func:`vaex.cache.get` simply look up the values in a global dict (``vaex.cache.cache``), but can be set for more complex behaviour.

A good library to use for in-memory caching is cachetools (https://pypi.org/project/cachetools/)

>>> import vaex
>>> import cachetools
>>> df = vaex.example()
>>> vaex.cache.cache = cachetools.LRUCache(1_000_000_000)  # 1gb cache

Configure using environment variables
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

See `Configuration <conf.html>`_ for more configuration options.

Especially when using the `vaex server <server.html>`_ it can be useful to turn on caching externally using enviroment variables.

    $ VAEX_CACHE=disk VAEX_CACHE_DISK_SIZE_LIMIT="10GB" python -m vaex.server

Will enable caching using :func:`vaex.cache.disk` and configure it to use at max 10 GB of disk space.

When using Vaex in combination with Flask or Plotly Dash, and using gunicorn for scaling, it can be useful
to use a multilevel cache, where the first cache is small but low latency (and private for each progress),
and a second higher latency disk cache that is shared among all processes.

    $ VAEX_CACHE="memory,disk" VAEX_CACHE_DISK_SIZE_LIMIT="10GB" VAEX_CACHE_MEMORY_SIZE_LIMIT="1GB" gunicorn -w 16 app:server

'''
import functools
from typing import ChainMap, MutableMapping
import logging
import shutil
import uuid
import os
import pickle
import threading


import dask.base
import pyarrow as pa
from vaex.promise import Promise

import vaex.utils
import vaex.settings

diskcache = vaex.utils.optional_import('diskcache')
_redis = vaex.utils.optional_import('redis')
cachetools = vaex.utils.optional_import('cachetools')

log = logging.getLogger('vaex.cache')

cache = None
# used for testing
_cache_hit = 0
_cache_miss = 0

class _cleanup:
    def __init__(self, gen, result):
        self._gen = gen
        self._result = result

    def __repr__(self):
        return '<cache restore context manager>'

    def __enter__(self):
        log.debug('entered')
        return self._result

    def __exit__(self, type=None, value=None, traceback=None):
        try:
            next(self._gen)
        except StopIteration:
            return False
        else:
            raise RuntimeError("generator didn't stop")

def _with_cleanup(f):
    '''Runs f directly, and the code after yield at exit'''
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        gen = f(*args, **kwargs)
        result = next(gen) # run up to the yield
        return _cleanup(gen, result)
    return wrapper


class MultiLevelCache(ChainMap):
    """Use multiple caches, where we assume the first is the fastest"""

    def __getitem__(self, key):
        for level, mapping in enumerate(self.maps):
            try:
                value = mapping[key]
                # write back to lower levels
                for i in range(level):
                    self.maps[i][key] = value
                return value
            except KeyError:
                pass
        return self.__missing__(key)

    def __setitem__(self, key, value):
        for cache in self.maps:
            cache[key] = value

    def __delitem__(self, key):
        for cache in self.maps:
            del cache[key]


@_with_cleanup
def multilevel_cache(*caches):
    """Sets a multilevel cache, where the first caches should be the fastest (low latency)

    This is useful when using multiple processes (requiring a persistent disk cache) in combination
    with a low latency memory cache.

    >>> import cachetools
    >>> import diskcache
    >>> l1 = cachetools.LRUCache(1_000_000_000)
    >>> l2 = diskcache.Cache()
    >>> vaex.cache.multilevel_cache(l1, l2)
    """
    global cache
    log.debug("set cache to multilevel")
    old_cache = cache
    cache = MultiLevelCache(*caches)
    yield
    log.debug("restore old cache")
    cache = old_cache


@_with_cleanup
def memory_infinite(clear=False):
    '''Sets a dict a cache, creating an infinite cache.
    
    Calling multiple times with clear=False will keep the current cache (useful in notebook usage)
    '''
    global cache
    log.debug("set cache to infinite")
    old_cache = cache
    if clear or (not isinstance(old_cache, dict)):
        cache = {}
    yield
    log.debug("restore old cache")
    cache = old_cache


@_with_cleanup
def memory(maxsize=vaex.settings.cache.memory_size_limit, classname="LRUCache", clear=False):
    """Sets a memory cache using cachetools (https://cachetools.readthedocs.io/).

    Calling multiple times with clear=False will keep the current cache (useful in notebook usage).

    :param int or str maxsize: Max size of cache in bytes (or use a string like '128MB')
    :param str classname: classname in the cachetools library used for the cache (e.g. LRUCache, MRUCache).
    :param bool clear: If False, will always set a new cache, when true, it will keep the cache when it is of the same type.
    """
    global cache
    from dask.utils import parse_bytes

    maxsize = parse_bytes(maxsize)
    log.debug("set cache to memory (cachetools)")
    old_cache = cache
    if isinstance(classname, str):
        cls = getattr(cachetools, classname)
    else:
        cls = classname
    if clear or type(cache) != cls:
        cache = cls(maxsize=maxsize)
    yield
    log.debug("restore old cache")
    cache = old_cache


@_with_cleanup
def disk(clear=False, size_limit=vaex.settings.cache.disk_size_limit, eviction_policy="least-recently-stored"):
    """Stored cached values using the diskcache library.

    See configuration details at `configuration of cache <conf.html#disk-size-limit>`_. and `configuration of paths <conf.html#cache-compute>`_

    :param int or str size_limit: Max size of cache in bytes (or use a string like '128MB')
        See http://www.grantjenks.com/docs/diskcache/tutorial.html?highlight=eviction#tutorial-settings for more details.
    :param str eviction_policy: Eviction policy,
        See http://www.grantjenks.com/docs/diskcache/tutorial.html?highlight=eviction#tutorial-eviction-policies
    :param bool clear: Remove all disk space used for caching before turning on cache.
    """
    from dask.utils import parse_bytes
    size_limit = parse_bytes(size_limit)
    global cache
    old_cache = cache
    path = vaex.settings.cache.path
    if clear:
        try:
            log.debug(f"Clearing disk cache: {path}")
            shutil.rmtree(path)
        except OSError:  # Windows wonkiness
            log.exception(f"Error clearing disk cache: {path}")
    log.debug(f"Initializing disk cache: {path}")
    cache = diskcache.Cache(path, size_limit=size_limit, eviction_policy=eviction_policy)
    yield
    log.debug("Restored old cache")
    cache = old_cache


class _RedisPickle(MutableMapping):
    '''Wraps a client such that the values are pickled/unpickled'''
    def __init__(self, client):
        self.client = client

    def __getitem__(self, key):
        return pickle.loads(self.client[key])

    def __setitem__(self, key, value):
        self.client[key] = pickle.dumps(value)

    def __delitem__(self, key):
        del self.client[key]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.client.keys())

    def keys(self):
        return self.client.keys()


@_with_cleanup
def redis(client=None):
    '''Uses Redis for caching.

    :param client: Redis client, if None, will call redis.Redis()
    '''
    global cache
    log.debug("Set cache to Redis")
    old_cache = cache
    if (not isinstance(old_cache, _redis.Redis)):
        if client is None:
            client = _redis.Redis()
        cache = _RedisPickle(client)
    yield
    log.debug("restore old cache")
    cache = old_cache


@_with_cleanup
def on(type="memory_infinite", **kwargs):
    if "," in type:
        cache_names = type.split(",")
        log.debug("Set multilevel cache to %r", cache_names)
        caches = []
        with off():  # so we don't change the cache global
            for name in cache_names:
                on(name)
                caches.append(cache)
        c = multilevel_cache(*caches)
    else:
        log.debug("Set cache to %r", type)
        if type == "memory":
            c = memory(**kwargs)
        elif type == "memory_infinite":
            c = memory_infinite(**kwargs)
        elif type == "disk":
            c = disk(**kwargs)
        elif type == "redis":
            c = redis(**kwargs)
        else:
            raise ValueError(f"Unknown type of cache {type}")
    yield
    c.__exit__()


@_with_cleanup
def off():
    '''Turns off caching, or temporary when used as context manager

    >>> import vaex
    >>> df = vaex.example()
    >>> vaex.cache.memory_infinite()  # cache on
    <cache restore context manager>
    >>> with vaex.cache.off():
    ...     df.x.sum()  # calculated without cache
    array(-20884.64307324)
    >>> df.x.sum()  # calculated with cache
    array(-20884.64307324)
    >>> vaex.cache.off()  # cache off
    <cache restore context manager>
    >>> df.x.sum()  # calculated without cache
    array(-20884.64307324)
    '''

    global cache
    old_cache = cache
    cache = None
    yield
    cache = old_cache


def is_on():
    '''Returns True when caching is enabled'''
    return cache is not None


def set(key, value, type=None, duration_wallclock=None):
    '''Set a cache value

    Useful to more advanced strategies, where we want to have different behaviour based on the type
    and costs. Implementations can set this function override the default behaviour:

    >>> import vaex
    >>> vaex.cache.memory_infinite()  # doctest: +SKIP
    >>> def my_smart_cache_setter(key, value, type=None, duration_wallclock=None):
    ...     if duration_wallclock >= 0.1:  # skip fast calculations
    ...         vaex.cache.cache[key] = value
    ...
    >>> vaex.cache.set = my_smart_cache_setter

    :param str key: key for caching
    :param type: Currently unused.
    :param float duration_wallclock: Time spend on calculating the result (in wallclock time).
    :param value: Any value, typically needs to be pickleable (unless stored in memory)
    '''
    cache[key] = value


def get(key, default=None, type=None):
    '''Looks up the cache value for the key, or returns the default

    Will return None if the cache is turned off.

    :param str key: Cache key.
    :param default: Return when cache is on, but key not in cache
    :param type: Currently unused.
    '''
    global _cache_miss, _cache_hit
    if is_on():
        value = cache.get(key, default)
        # this might not be the best way to check for a cache hit or miss
        # but it is sufficient for testing
        if value is default:
            _cache_miss += 1
        else:
            _cache_hit += 1
        return value


class _ThreadLocalCallablePatch:
    def __init__(self, original, patch):
        self.local = threading.local()
        self.local.patch = patch
        self.original = original

    def __call__(self, *args, **kwargs):
        f = getattr(self.local, 'patch', self.original)
        return f(*args, **kwargs)


def _explain(*args):
    raise TypeError('You have passed in an object for which we cannot determine a fingerprint')


def fingerprint(*args, **kwargs):
    try:
        original = uuid.uuid4
        if isinstance(original, _ThreadLocalCallablePatch):
            original = original.original
        uuid.uuid4 = _ThreadLocalCallablePatch(original, _explain)
        try:
            return dask.base.tokenize(*args, **kwargs)
        except TypeError as e:
            raise TypeError(f"Could not calculate fingerprint for {args}/{kwargs}") from e
    finally:
        uuid.uuid4 = original


def output_file(callable=None, path_input=None, fs_options_input={}, fs_input=None, path_output=None, fs_options_output={}, fs_output=None):
    """Decorator to do cached conversion from path_input → path_output.

    Caching is active if with the input file, the *args, and **kwargs are unchanged, otherwise callable
    will be called again.


    """
    def wrapper1(callable=callable):
        def wrapper2(callable=callable):
            def call(callable=callable, *args, **kwargs):
                base, ext, fs_options_output_ = vaex.file.split_ext(path_output, fs_options_output)
                path_output_meta = base + '.yaml'
                # this fingerprint changes if input changes, or args, or kwargs
                fp = fingerprint(vaex.file.fingerprint(path_input, fs_options_input, fs=fs_input), args, kwargs)

                def write_fingerprint():
                    log.info('saving fingerprint %r for %s → %s conversion to %s', fp, path_input, path_output, path_output_meta)
                    with vaex.file.open(path_output_meta, 'w', fs_options=fs_options_output_, fs=fs_output) as f:
                        f.write(f"# this file exists so that we know when not to do the\n# {path_input} → {path_output} conversion\n")
                        vaex.utils.yaml_dump(f, {'fingerprint': fp})

                lockname = os.path.join('convert', f"{fp}.lock")
                with vaex.utils.file_lock(lockname):
                    if not vaex.file.exists(path_output, fs_options=fs_options_output_, fs=fs_output):
                        log.info('file %s does not exist yet, running conversion %s → %s', path_output_meta, path_input, path_output)
                        value = callable(*args, **kwargs)
                        write_fingerprint()
                        return value


                    if not vaex.file.exists(path_output_meta, fs_options=fs_options_output_, fs=fs_output):
                        log.info('file including fingerprint not found (%) or does not exist yet, running conversion %s → %s', path_output_meta, path_input, path_output)
                        value = callable(*args, **kwargs)
                        write_fingerprint()
                        return value

                    # load fingerprint
                    with vaex.file.open(path_output_meta, fs_options=fs_options_output_, fs=fs_output) as f:
                        output_meta = vaex.utils.yaml_load(f)

                    if output_meta is None or output_meta['fingerprint'] != fp:
                        if output_meta is None:
                            log.error('fingerprint for %s (%s) was empty', path_input, path_output_meta)
                        else:
                            log.info('fingerprint for %s is out of date, rerunning conversion to %s', path_input, path_output)
                        value = callable(*args, **kwargs)
                        write_fingerprint()
                        return value
                    else:
                        log.info('fingerprint for %s did not change, reusing converted file %s', path_input, path_output)
            return call
        return wrapper2
    return wrapper1()

dask.base.normalize_token.register(pa.DataType, repr)


@dask.base.normalize_token.register(pa.Array)
def _normalize(ar):
    return vaex.dataset.hash_array_data(ar)


def _memoize(f=None, key_function=None, type='computed', delay=False):
    import time
    def wrapper_top(f=f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if vaex.cache.is_on():
                key = key_function()
                value = vaex.cache.get(key, type=type)
                if value is None:
                    t0 = time.time()
                    if delay:
                        promise = f(*args, **kwargs)
                        def set(value):
                            duration_wallclock = time.time() - t0
                            vaex.cache.set(key, value, type='computed', duration_wallclock=duration_wallclock)
                            return value
                        return promise.then(set)
                    else:
                        value = f(*args, **kwargs)
                        duration_wallclock = time.time() - t0
                        vaex.cache.set(key, value, type='computed', duration_wallclock=duration_wallclock)
                        return value
                else:
                    if delay:
                        return Promise.fulfilled(value)
                    else:
                        return value
            else:
                return f(*args, **kwargs)
        return wrapper
    if f is None:
        return wrapper_top
    else:
        return wrapper_top()


if vaex.settings.cache.type:
    on(vaex.settings.cache.type)



__author__ = 'maartenbreddels'
import PIL.Image
import PIL.ImageDraw
try:
    from StringIO import StringIO
    py3 = False
except ImportError:
    from io import StringIO, BytesIO
    py3 = True
import collections
import numpy as np
import matplotlib.colors
import warnings
from base64 import b64encode


def rgba_2_pil(rgba):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        im = PIL.Image.fromarray(rgba[::-1], "RGBA")  # , "RGBA", 0, -1)
    return im


def pil_2_data(im, format="png"):
    if py3:  # py3 case
        f = BytesIO()
    else:
        f = StringIO()
    im.save(f, format)
    return f.getvalue()


def rgba_to_url(rgba):
    bit8 = rgba.dtype == np.uint8
    if not bit8:
        rgba = (rgba * 255.).astype(np.uint8)
    im = rgba_2_pil(rgba)
    data = pil_2_data(im)
    data = b64encode(data)
    if py3:
        data = data.decode("ascii")
    imgurl = "data:image/png;base64," + data + ""
    return imgurl


pdf_modes = collections.OrderedDict()
pdf_modes["multiply"] = np.multiply
pdf_modes["screen"] = lambda a, b: a + b - a * b
pdf_modes["darken"] = np.minimum
pdf_modes["lighten"] = np.maximum

cairo_modes = collections.OrderedDict()
cairo_modes["saturate"] = (lambda aA, aB: np.minimum(1, aA + aB),
                           lambda aA, xA, aB, xB, aR: (np.minimum(aA, 1 - aB) * xA + aB * xB) / aR)
cairo_modes["add"] = (lambda aA, aB: np.minimum(1, aA + aB),
                      lambda aA, xA, aB, xB, aR: (aA * xA + aB * xB) / aR)

modes = list(pdf_modes.keys()) + list(cairo_modes.keys())


def background(shape, color="white", alpha=1, bit8=True):
    rgba = np.zeros(shape + (4,))
    rgba[:] = np.array(matplotlib.colors.colorConverter.to_rgba(color))
    rgba[..., 3] = alpha
    if bit8:
        return (rgba * 255).astype(np.uint8)
    else:
        return rgba


def fade(image_list, opacity=0.5, blend_mode="multiply"):
    result = image_list[0] * 1.
    for i in range(1, len(image_list)):
        result[result[..., 3] > 0, 3] = 1
        layer = image_list[i] * 1.0
        layer[layer[..., 3] > 0, 3] = opacity
        result = blend([result, layer], blend_mode=blend_mode)
    return result


def blend(image_list, blend_mode="multiply"):
    bit8 = image_list[0].dtype == np.uint8
    image_list = image_list[::-1]
    rgba_dest = image_list[0] * 1
    if bit8:
        rgba_dest = (rgba_dest / 255.).astype(np.float)
    for i in range(1, len(image_list)):
        rgba_source = image_list[i]
        if bit8:
            rgba_source = (rgba_source / 255.).astype(np.float)
        # assert rgba_source.dtype == image_list[0].dtype, "images have different types: first has %r, %d has %r" % (image_list[0].dtype, i, rgba_source.dtype)

        aA = rgba_source[:, :, 3]
        aB = rgba_dest[:, :, 3]

        if blend_mode in pdf_modes:
            aR = aA + aB * (1 - aA)
        else:
            aR = cairo_modes[blend_mode][0](aA, aB)
        mask = aR > 0
        for c in range(3):  # for r, g and b
            xA = rgba_source[..., c]
            xB = rgba_dest[..., c]
            if blend_mode in pdf_modes:
                f = pdf_modes[blend_mode](xB, xA)
                with np.errstate(divide='ignore', invalid='ignore'):  # these are fine, we are ok with nan's in vaex
                    result = ((1. - aB) * aA * xA + (1. - aA) * aB * xB + aA * aB * f) / aR
            else:
                result = cairo_modes[blend_mode][1](aA, xA, aB, xB, aR)
            with np.errstate(divide='ignore', invalid='ignore'):  # these are fine, we are ok with nan's in vaex
                result = (np.minimum(aA, 1 - aB) * xA + aB * xB) / aR
            # print(result)
            rgba_dest[:, :, c][(mask, )] = np.clip(result[(mask,)], 0, 1)
        rgba_dest[:, :, 3] = np.clip(aR, 0., 1)
    rgba = rgba_dest
    if bit8:
        rgba = (rgba * 255).astype(np.uint8)
    return rgba


def monochrome(I, color, vmin=None, vmax=None):
    """Turns a intensity array to a monochrome 'image' by replacing each intensity by a scaled 'color'

    Values in I between vmin  and vmax get scaled between 0 and 1, and values outside this range are clipped to this.

    Example
    >>> I = np.arange(16.).reshape(4,4)
    >>> color = (0, 0, 1) # red
    >>> rgb = vx.image.monochrome(I, color) # shape is (4,4,3)

    :param I: ndarray of any shape (2d for image)
    :param color: sequence of a (r, g and b) value
    :param vmin: normalization minimum for I, or np.nanmin(I) when None
    :param vmax: normalization maximum for I, or np.nanmax(I) when None
    :return:
    """
    if vmin is None:
        vmin = np.nanmin(I)
    if vmax is None:
        vmax = np.nanmax(I)
    normalized = (I - vmin) / (vmax - vmin)
    return np.clip(normalized[..., np.newaxis], 0, 1) * np.array(color)


def polychrome(I, colors, vmin=None, vmax=None, axis=-1):
    """Similar to monochrome, but now do it for multiple colors

    Example
    >>> I = np.arange(32.).reshape(4,4,2)
    >>> colors = [(0, 0, 1), (0, 1, 0)] # red and green
    >>> rgb = vx.image.polychrome(I, colors) # shape is (4,4,3)

    :param I: ndarray of any shape (3d will result in a 2d image)
    :param colors: sequence of [(r,g,b), ...] values
    :param vmin: normalization minimum for I, or np.nanmin(I) when None
    :param vmax: normalization maximum for I, or np.nanmax(I) when None
    :param axis: axis which to sum over, by default the last
    :return:
    """
    axes_length = len(I.shape)
    allaxes = list(range(axes_length))
    otheraxes = list(allaxes)
    otheraxes.remove((axis + axes_length) % axes_length)
    otheraxes = tuple(otheraxes)

    if vmin is None:
        vmin = np.nanmin(I, axis=otheraxes)
    if vmax is None:
        vmax = np.nanmax(I, axis=otheraxes)
    normalized = (I - vmin) / (vmax - vmin)
    return np.clip(normalized, 0, 1).dot(colors)

# c_r + c_g + c_b



packages = [
    "vaex",
    "vaex-core",
    "vaex-viz",
    "vaex-hdf5",
    "vaex-server",
    "vaex-astro",
    "vaex-ui",
    "vaex-jupyter",
    "vaex-ml",
    "vaex-arrow",
    "vaex-graphql",
    "vaex-enterprise",
]


# Attempt to use importlib.metadata first because it's much faster
# though it's only available in Python 3.8+ so we'll need to fall
# back to pkg_resources for Python 3.7 support
# See https://github.com/jacob-indigo/geoalchemy2/blob/c7351b2086d5d711cbeb0c92f27002829e72bab3/geoalchemy2/__init__.py#L446
def get_version(package):
    try:
        import importlib.metadata
    except ImportError:
        from pkg_resources import DistributionNotFound
        from pkg_resources import get_distribution

        try:
            version = get_distribution(package).version
        except DistributionNotFound:  # pragma: no cover
            pass
    else:
        try:
            version = importlib.metadata.version(package)
        except importlib.metadata.PackageNotFoundError:  # pragma: no cover
            pass
    return version


def get_versions():
    installed_packages = {}
    for p in packages:
        try:
            installed_packages[p] = get_version(p)
        except Exception:
            pass
    return installed_packages
